from .scales import Scale
from .scales import major, major_triad, major_intervals
from .scales import minor, minor_triad, minor_intervals
from .scales import diminished_seventh, dominant_seventh, half_diminished_seventh, minor_seventh, major_seventh, minor_major_seventh

from .scales import chords

from .note import NoteAccidental, NotePitchClasses, NotePitchClassesInv, NotePitchClassesIndexes, Note
from .chord import Chord
from .interval import  Interval


from musikla.parser.abstract_syntax_tree.statements.while_loop_node import WhileLoopStatementNode
from typing import Any, Dict, List, Optional, Tuple, cast
from musikla.core import Context, Value
from ..node import Node, ValueNode
from ..expressions import FunctionExpressionNode, StringLiteralNode, BoolLiteralNode, ListComprehensionNode, VariableExpressionNode, ArrayLiteralNode, BlockNode
from ..statements import ForLoopStatementNode, IfStatementNode, StatementsListNode, VariableDeclarationStatementNode
from .macro import MacroNode

ModifierNames = [ 'hold', 'extend', 'toggle', 'repeat', 'release' ]

def handle_modifiers ( modifiers : List[str] ) -> Tuple[Dict[str, bool], Optional[str]]:
    props = dict()

    rest = None

    for i in range( len( modifiers ) - 1, -1, -1 ):
        if modifiers[ i ] in ModifierNames:
            props[ modifiers[ i ] ] = True
        else:
            rest = '+'.join( modifiers[ 0:i + 1 ] )
            break

    return (props, rest)

class KeyboardShortcutMacroNode(MacroNode):
    def __init__ ( self, modifiers : List[str], args : List[str], expression : Node, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.modifiers : List[str] = modifiers
        self.args : List[str] = args
        self.expression : Node = expression

        (kargs_raw, shortcut) = handle_modifiers( modifiers )

        kargs : Dict[str, Node] = dict( ( m, BoolLiteralNode( v ) ) for m, v in kargs_raw.items() )

        if self.args:
            kargs[ 'args' ] = ValueNode( self.args )

        self.virtual_node : Node = FunctionExpressionNode(
            VariableExpressionNode( "keyboard\\register" ),
            [ None, StringLiteralNode( shortcut ), expression ],
            kargs,
            position = self.position
        )

    def set_keyboard ( self, keyboard : Node ):
        cast( FunctionExpressionNode, self.virtual_node ).parameters[ 0 ] = keyboard

class KeyboardShortcutDynamicMacroNode(MacroNode):
    def __init__ ( self, shortcut : Node, modifiers : List[str], args : List[str], expression : Node, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.shortcut : Node = shortcut
        self.modifiers : List[str] = modifiers
        self.args : List[str] = args
        self.expression : Node = expression

        (kargs_raw, rest) = handle_modifiers( modifiers )

        kargs : Dict[str, Node] = dict( ( m, BoolLiteralNode( v ) ) for m, v in kargs_raw.items() )
        
        if self.args:
            kargs[ 'args' ] = ValueNode( self.args )

        if rest != None:
            raise BaseException( "Keyboard shortcut with invalid modifiers: %s" % rest )

        self.virtual_node : Node = FunctionExpressionNode( 
            VariableExpressionNode( "keyboard\\register" ), 
            [ None, shortcut, expression ], 
            kargs, 
            position = self.position
        )

    def set_keyboard ( self, keyboard : Node ):
        cast( FunctionExpressionNode, self.virtual_node ).parameters[ 0 ] = keyboard

class KeyboardShortcutComprehensionMacroNode(MacroNode):
    def __init__ ( self, comprehension : ListComprehensionNode, modifiers : List[str], args : List[str], expression : Node, position : Tuple[int, int, int] = None ):
        super().__init__( position )
        
        self.comprehension : ListComprehensionNode = comprehension
        self.modifiers : List[str] = modifiers
        self.args : List[str] = args
        self.expression : Node = expression

        (kargs_raw, rest) = handle_modifiers( modifiers )

        kargs : Dict[str, Node] = dict( ( m, BoolLiteralNode( v ) ) for m, v in kargs_raw.items() )

        if rest != None:
            raise BaseException( "Keyboard shortcut with invalid modifiers: %s" % rest )

        if self.args:
            kargs[ 'args' ] = ValueNode( self.args )

        node = FunctionExpressionNode( 
            VariableExpressionNode( "keyboard\\register" ), 
            [ None, self.comprehension.expression, expression ], 
            kargs
        )

        if self.comprehension.condition != None:
            node = IfStatementNode( self.comprehension.condition, node )

        r = FunctionExpressionNode( VariableExpressionNode( "range" ), [ self.comprehension.min, self.comprehension.max ] )

        self.virtual_node : Node = ForLoopStatementNode( 
            self.comprehension.variable, 
            r,
            node, 
            position = position 
        )

    def set_keyboard ( self, keyboard : Node ):
        node = cast( ForLoopStatementNode, self.virtual_node ).body

        if isinstance( node, IfStatementNode ):
            node = node.body
        
        cast( FunctionExpressionNode, node ).parameters[ 0 ] = keyboard

class KeyboardForLoopMacroNode( MacroNode ):
    def __init__ ( self, variable : str, iterator : Node, shortcuts : List[MacroNode], position : Tuple[int, int, int] = None ):
        super().__init__( position )
    
        self.variable : str = variable
        self.iterator : Node = iterator
        self.shortcuts : List[MacroNode] = shortcuts

        body = StatementsListNode( [ macro.virtual_node for macro in self.shortcuts if macro.virtual_node is not None ] )

        self.virtual_node = ForLoopStatementNode( self.variable, self.iterator, body, position = position )

    def set_keyboard ( self, keyboard : Node ):
        for macro in self.shortcuts:
            cast( Any, macro ).set_keyboard( keyboard )

class KeyboardWhileLoopMacroNode( MacroNode ):
    def __init__ ( self, condition : Node, shortcuts : List[MacroNode], position : Tuple[int, int, int] = None ):
        super().__init__( position )
    
        self.condition : Node = condition
        self.shortcuts : List[MacroNode] = shortcuts

        body = StatementsListNode( [ macro.virtual_node for macro in self.shortcuts if macro.virtual_node is not None ] )

        self.virtual_node = WhileLoopStatementNode( self.condition, body, position = position )

    def set_keyboard ( self, keyboard : Node ):
        for macro in self.shortcuts:
            cast( Any, macro ).set_keyboard( keyboard )

class KeyboardIfMacroNode( MacroNode ):
    def __init__ ( self, condition : Node, shortcuts : List[MacroNode], else_shortcuts : List[MacroNode] = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )
    
        self.condition : Node = condition
        self.shortcuts : List[MacroNode] = shortcuts
        self.else_shortcuts : List[MacroNode] = else_shortcuts

        body = StatementsListNode( [ macro.virtual_node for macro in self.shortcuts if macro.virtual_node is not None ] )
        else_body = StatementsListNode( [ macro.virtual_node for macro in self.else_shortcuts if macro.virtual_node is not None ] ) if else_shortcuts is not None else None

        self.virtual_node = IfStatementNode( self.condition, body, else_body, position = position )

    def set_keyboard ( self, keyboard : Node ):
        for macro in self.shortcuts:
            cast( Any, macro ).set_keyboard( keyboard )

        if self.else_shortcuts is not None:
            for macro in self.else_shortcuts:
                cast( Any, macro ).set_keyboard( keyboard )

class KeyboardBlockMacroNode( MacroNode ):
    def __init__ ( self, body : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )
    
        self.body : Optional[Node] = body

        self.virtual_node = self.body # StatementsListNode( [ macro.virtual_node for macro in self.body if macro.virtual_node is not None ] )

    def set_keyboard ( self, keyboard : Node ):
        pass

class KeyboardDeclarationMacroNode(MacroNode):
    def __init__ ( self, shortcuts : List[Node], flags : List[str] = None, prefix : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        var_name = '__keyboard'

        # Clone the list so that our changes don't affect the original list
        shortcuts = list( shortcuts )

        for node in shortcuts:
            cast( Any, node ).set_keyboard( VariableExpressionNode( var_name ) )

        if flags:
            shortcuts.insert( 0, FunctionExpressionNode( VariableExpressionNode( "keyboard\\push_flags" ), cast( Any, [ VariableExpressionNode( var_name ) ] ) + [ StringLiteralNode( f ) for f in flags ] ) )
            shortcuts.append( FunctionExpressionNode( VariableExpressionNode( "keyboard\\pop_flags" ), cast( Any, [ VariableExpressionNode( var_name ) ] ) + [ StringLiteralNode( f ) for f in flags ] ) )

        if prefix:
            shortcuts.insert( 0, FunctionExpressionNode( VariableExpressionNode( "keyboard\\push_prefix" ), cast( Any, [ VariableExpressionNode( var_name ) ] ) + [ prefix ] ) )
            shortcuts.append( FunctionExpressionNode( VariableExpressionNode( "keyboard\\pop_prefix" ), cast( Any, [ VariableExpressionNode( var_name ) ] ) ) )

        shortcuts.insert( 0, VariableDeclarationStatementNode( var_name, FunctionExpressionNode( VariableExpressionNode( "keyboard\\create" ) ), local = True ) )
        shortcuts.append( VariableExpressionNode( var_name ) )

        self.virtual_node = BlockNode( StatementsListNode( shortcuts, position ) )

        # from musikla.parser import CodePrinter
        
        # print( CodePrinter().print( self.virtual_node ) )



from .transformer import Transformer
from musikla.core import MusicBuffer
from ..note import NoteEvent, NoteOnEvent, NoteOffEvent
from ..chord import ChordEvent, ChordOnEvent, ChordOffEvent
from collections import defaultdict
from typing import Callable

class DecomposeNotesTransformer( Transformer ):
    def transform ( self ):
        offs : MusicBuffer = MusicBuffer()

        while True:
            done, event = yield

            if done: break

            for ev in offs.collect( event.timestamp ):
                self.add_output( ev )
            
            if isinstance( event, NoteEvent ):
                offs.append( event.note_off )

                self.add_output( event.note_on )
            elif isinstance( event, ChordEvent ):
                offs.append( event.chord_off )

                self.add_output( event.chord_on )
            else:
                self.add_output( event )

        for event in offs.collect():
            self.add_output( event )


from musikla.parser.printer import CodePrinter
from typing import Optional, Tuple
from musikla.core import Context
from ..node import Node
from .statement_node import StatementNode

class IfStatementNode( StatementNode ):
    def __init__ ( self, condition : Node, body : Node, else_body : Node = None, position : Tuple[int, int, int] = None ):
        from ..expressions.block_node import BlockNode

        super().__init__( position )

        self.condition : Node = condition
        self.body : Node = body
        self.else_body : Optional[Node] = else_body

        if isinstance( self.body, BlockNode ):
            self.body.fork_context = False
            self.body.create_stack_frame = False

        if isinstance( self.else_body, BlockNode ):
            self.else_body.fork_context = False
            self.else_body.create_stack_frame = False

    def __eval__ ( self, context : Context ):
        condition_value = self.condition.eval( context )

        result = None

        if condition_value:
            result = self.body.eval( context )
        elif self.else_body != None:
            result = self.else_body.eval( context )

        return result

    def to_source ( self, printer : CodePrinter ):
        from ..expressions.group_node import GroupNode
        from ..expressions.block_node import BlockNode

        printer.add_token( 'if ' )

        self.condition.to_source( printer )

        if not isinstance( self.condition, GroupNode ) and not isinstance( self.body, BlockNode ):
            printer.add_token( " then " )
        else:
            printer.add_token( " " )

        self.body.to_source( printer )

        if self.else_body is not None:
            printer.add_token( " else " )

            self.else_body.to_source( printer )



from .expression_node import ExpressionNode
from .variable_expression_node import VariableExpressionNode
from .function_expression_node import FunctionExpressionNode, FunctionChainExpressionNode
from .group_node import GroupNode
from .string_literal_node import StringLiteralNode
from .number_literal_node import NumberLiteralNode
from .property_accessor_node import PropertyAccessorNode
from .bool_literal_node import BoolLiteralNode
from .constant_node import ConstantNode
from .none_literal_node import NoneLiteralNode
from .array_literal_node import ArrayLiteralNode
from .object_literal_node import ObjectLiteralNode, Object
from .binary_operator_node import BinaryOperatorNode, PlusBinaryOperatorNode, MinusBinaryOperatorNode, PowBinaryOperatorNode, MultBinaryOperatorNode, DivBinaryOperatorNode
from .binary_operator_node import AndLogicOperatorNode, OrLogicOperatorNode
from .binary_operator_node import GreaterComparisonOperatorNode, GreaterEqualComparisonOperatorNode
from .binary_operator_node import EqualComparisonOperatorNode, NotEqualComparisonOperatorNode
from .binary_operator_node import LesserComparisonOperatorNode, LesserEqualComparisonOperatorNode
from .binary_operator_node import IsComparisonOperatorNode, IsNotComparisonOperatorNode
from .binary_operator_node import InComparisonOperatorNode, NotInComparisonOperatorNode
from .unary_operator_node import UnaryOperatorNode, NotOperatorNode
from .block_node import BlockNode
from .list_comprehension_node import ListComprehensionNode



from typing import List
from .transformer import Transformer
from ..note import NoteEvent

class TerminationMelodyTransformer(Transformer):
    def __init__ ( self, events, exclude_matches : bool = True, stop_on_match : bool = True ):
        super().__init__()

        self.pattern = [ ev for ev in events if isinstance( ev, NoteEvent ) ]
        self.exclude_matches : bool = True
        self.stop_on_match : bool = stop_on_match
        self.matching_now : bool = False
        self.matched : bool = False

    def transform ( self ):
        buffer = []

        while True:
            done, event = yield

            if done: break

            if isinstance( event, NoteEvent ):
                buffer.append( BufferSlot( event ) )

                while len( buffer ) > len( self.pattern ):
                    for ev in buffer.pop( 0 ): self.add_output( ev )

                if len( buffer ) == len( self.pattern ):
                    matched = True

                    for i in range( len( self.pattern ) ):
                        if int( self.pattern[ i ] ) != int( buffer[ i ].event ):
                            matched = False

                            break
                    
                    self.matching_now = matched

                    if matched:
                        self.matched = True
                        
                        if not self.exclude_matches:
                            for slot in buffer:
                                for ev in slot:
                                    self.add_output( ev )
                        
                            buffer = []
                        
                        
                        if self.stop_on_match:
                            buffer = []

                            self.end_output()
            elif buffer:
                buffer[ -1 ].append( event )
            else:
                self.add_output( event )
            
        for slot in buffer:
            for ev in slot:
                self.add_output( ev )

class BufferSlot:
    def __init__ ( self, event ):
        self.event = event
        self.next_events : List = None
    
    def append ( self, event ):
        if not self.next_events:
            self.next_events = []

        self.next_events.append( event )

    def __iter__ ( self ):
        yield self.event

        if self.next_events:
            for event in self.next_events:
                yield event



from typing import Optional
from .transformer import Transformer
from ..note import NoteOnEvent, NoteOffEvent
from ..chord import ChordOnEvent, ChordOffEvent
from typing import Callable, List, Tuple

class BalanceNotesTransformer( Transformer ):
    def __init__ ( self, get_time : Callable = None ):
        super().__init__()

        self.get_time : Optional[Callable] = get_time

    def transform ( self ):
        tracker = NoteTracker()

        last_time : int = 0

        while True:
            done, value = yield

            if done: break

            last_time = max( value.end_timestamp, last_time )

            tracker.process( value )

            self.add_output( value )

        for event in tracker.close( last_time if self.get_time is None else self.get_time() ):
            self.add_output( event )

class NoteTracker:
    def __init__ ( self ):
        self.active_notes : List[NoteOnEvent] = []
        self.active_chords : List[ChordOnEvent] = []

    def activate ( self, note_on : NoteOnEvent ):
        self.active_notes.append( note_on )

    def deactivate ( self, note_off : NoteOffEvent ):
        match_index = None

        for i, on in enumerate( self.active_notes ):
            if on.voice.name == note_off.voice.name and int( on ) == int( note_off ):
                match_index = i

        if match_index != None:
            del self.active_notes[ match_index ]

    def activate_chord ( self, chord_on : ChordOnEvent ):
        self.active_chords.append( chord_on )

    def deactivate_chord ( self, chord_off : ChordOffEvent ):
        match_index = None

        for i, on in enumerate( self.active_chords ):
            if on.voice.name == chord_off.voice.name and on.pitches == chord_off.pitches:
                match_index = i

        if match_index != None:
            del self.active_chords[ match_index ]

    def process ( self, event ):
        if isinstance( event, NoteOnEvent ):
            self.activate( event )
        elif isinstance( event, NoteOffEvent ):
            self.deactivate( event )
        elif isinstance( event, ChordOnEvent ):
            self.activate_chord( event )
        elif isinstance( event, ChordOffEvent ):
            self.deactivate_chord( event )

    def close ( self, time : int ):
        for on in self.active_notes:
            yield on.note_off( time )
        
        for on in self.active_chords:
            yield on.chord_off( time )



from sys import base_exec_prefix
from typing import Any, Optional, Tuple
from typeguard import check_type
from .music import Music, MusicGen

class Value:
    @staticmethod
    def assignment ( value ):
        # TODO
        if isinstance( value, Music ):
            return value.shared()
        
        return value

    @staticmethod
    def forked ( value ):
        if isinstance( value, Music ):
            def _gen ( context ):
                for ev in value.expand( context ):
                    context.cursor = max( ev.end_timestamp, context.cursor )

                    yield ev

            return MusicGen( None, _gen )
        else:
            return value

    @staticmethod
    def expect ( value, typehint, name : str = "", soft : bool = False ) -> bool:
        if soft:
            try:
                check_type( name, value, typehint )

                return True
            except:
                return False
        else:
            check_type( name, value, typehint )
            
            return True

    @staticmethod
    def typeof ( value ):
        return type( value )

    @staticmethod
    def eval ( context, node ) -> Any:
        if node == None: 
            return None

        return node.eval( context )

class CallableValue:
    def __init__ ( self, fn ):
        self.fn = fn

    def __call__ ( self, context, args, kargs ):
        return self.fn( context, *args, **kargs )

    def raw ( self ):
        return self.fn



from musikla.parser.printer import CodePrinter
from .expression_node import ExpressionNode
from typing import List, Tuple
from musikla.core import Value

class ArrayLiteralNode( ExpressionNode ):
    def __init__ ( self, values : List[ExpressionNode], position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.values : List[ExpressionNode] = values

    def __eval__ ( self, context ):
        return [ Value.assignment( Value.eval( context.fork(), node ) ) for node in self.values ]

    def to_source ( self, printer : CodePrinter ):
        with printer.block( '@[', ']' ):
            for i in range( len( self.values ) ):
                if i > 0:
                    printer.add_token( '; ' )
                
                self.values[ i ].to_source( printer )



from typing import Dict

class SharedContext():
    def __init__ ( self ):
        self.channel_count = 24
        self.channels = dict()
        self.release_channels = True

    @property
    def available_channels ( self ):
        return filter( lambda i: i not in self.channels, range( 1, self.channel_count ) )

    def register_instrument ( self, instrument ):
        channel = next( self.available_channels, None )

        if channel == None:
            raise BaseException( "No channel available found" )

        self.channels[ channel ] = instrument

        instrument.channel = channel

    def unregister_instrument ( self, instrument ):
        if self.release_channels:
            del self.channels[ instrument.channel ]

            instrument.channel = None



from .node import Node, MusiklaExecutionError
from .chord_node import ChordNode
from .music_node import MusicNode
from .music_parallel_node import MusicParallelNode
from .music_sequence_node import MusicSequenceNode
from .note_node import NoteNode
from .python_node import PythonNode
from .rest_node import RestNode
from .stack_frame_node import StackFrameNode


from .voice import Voice
from .shared_context import SharedContext
from .symbols_scope import SymbolsScope
from typing import Any, Hashable, Optional, cast, TYPE_CHECKING
from fractions import Fraction


if TYPE_CHECKING:
    from musikla.script import Script

class Library:
    def __init__ ( self, namespace : str = None ):
        self.namespace : Optional[str] = namespace
        self.context : Context = cast( Any, None )

    def on_link ( self, script ):
        pass

    def resolve ( self, name : str ) -> str:
        if self.namespace != None and self.namespace != '':
            if self.namespace.endswith( '\\' ) and name.startswith( '\\' ):
                return self.namespace + name[ 1: ]
            elif self.namespace.endswith( '\\' ) or name.startswith( '\\' ):
                return self.namespace + name
            else:
                return self.namespace + '\\' + name
        else:
            return name

    def lookup ( self, name : Hashable, container : str = "", recursive : bool = True, follow_pointers : bool = True, default = None ):
        return self.context.symbols.lookup( self.resolve( name ), container = container, recursive = recursive, follow_pointers = follow_pointers, default = default )

    def assign ( self, name : Hashable, value, container = "", follow_pointers : bool = True ):
        return self.context.symbols.assign( self.resolve( name ), value, container, follow_pointers )

    def lookup_instrument ( self, name ):
        return self.context.symbols.lookup_instrument( self.resolve( name ) )

    def assign_instrument ( self, instrument ):
        self.assign( instrument.name, instrument, container = "instruments" )
        
        return instrument

    def lookup_internal ( self, name ):
        return self.context.symbols.lookup_internal( self.resolve( name ) )

    def assign_internal ( self, name, value ):
        self.context.symbols.assign_internal( self.resolve( name ), value )
    
    def eval_file ( self, script, file ):
        script.execute_file( file, context = self.context, fork = False, silent = True )
    
    def eval ( self, script, code ):
        script.execute( code, context = self.context, fork = False, silent = True )

class Context():
    default : 'Context' = None

    @staticmethod
    def create ():
        ctx = Context(
            # voice = Voice( "default", instrument = Instrument( 'Acoustic Grand Piano', GeneralMidi.AcousticGrandPiano ) )
        )

        ctx.symbols.assign( 'is_prelude', True, container = 'internal' )
        ctx.symbols.assign( ctx.voice.name, ctx.voice )

        return ctx

    def __init__ ( self, 
                   shared : SharedContext = None, 
                   voice : Voice = None,
                   cursor : int = 0,
                   symbols : SymbolsScope = None,
                 ):
        self.shared : SharedContext = shared or SharedContext()
        self.voice : Optional[Voice] = voice or Voice()
        self.cursor : int = cursor
        self.symbols : SymbolsScope = symbols or SymbolsScope()

    @property
    def script ( self ) -> 'Script':
        return self.symbols.lookup( 'script', container = 'internal' )

    def fork ( self, cursor : int = None, symbols : SymbolsScope = None ) -> 'Context':
        return Context(
            shared = self.shared,
            voice = self.voice,
            cursor = self.cursor if cursor == None else cursor,
            symbols = self.symbols if symbols == None else symbols
        )

    def join ( self, *child_context ):
        for context in child_context:
            if type(context) == int:
                if context > self.cursor:
                    self.cursor = context
            else:
                if context.cursor > self.cursor:
                    self.cursor = context.cursor
    
    def get_value ( self, value : float ) -> float:
        return self.voice.get_value( value )

    def get_duration_ratio ( self ) -> float:
        return self.voice.get_duration_ratio()

    def get_duration ( self, value : float = None ) -> int:
        """Transform a not value into the real world milliseconds it takes, according to the voice's tempo and time signature"""
        return self.voice.get_duration( value )

    def from_duration ( self, milliseconds : int ) -> Fraction:
        """Transform a duration in milliseconds to an approximated note value relative to the tempo and time signature"""
        return self.voice.from_duration( milliseconds )

    def is_linked ( self, library : Library ) -> bool:
        return self.symbols.lookup( library.__class__, container = 'libraries' ) != None

    def link ( self, library : Library, script = None ):
        if not self.is_linked( library ):
            library.context = self

            self.symbols.assign( library.__class__, library, container = "libraries" )

            library.on_link( script )

    def library ( self, library ) -> Library:
        return self.symbols.lookup( library, container = "libraries" )

Context.default = Context.create()

class StackFrame:
    def __init__ ( self ):
        self.returned : bool = False
        self.returned_value : Any = None

    def ret ( self, value : Any = None ):
        self.returned = True
        self.returned_value = value


from typing import Tuple
from musikla.parser.printer import CodePrinter
from .expression_node import ExpressionNode

class BoolLiteralNode( ExpressionNode ):
    def __init__ ( self, value, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.value = value

    def __eval__ ( self, context ):
        return self.value

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'true' if self.value else 'false' )



from .balance_notes import BalanceNotesTransformer, NoteTracker
from .compose_notes import ComposeNotesTransformer
from .decompose_notes import DecomposeNotesTransformer
from .termination_melody import TerminationMelodyTransformer
from .sort import SortTransformer
from .tee import TeeTransformer
from .voice_identifier import VoiceIdentifierTransformer
from .notation_builder import NotationBuilderTransformer
from .decompose_chords import DecomposeChordsTransformer
from .compose_chords import ComposeChordsTransformer
from .annotate import AnnotateTransformer
from .ensure_order import EnsureOrderTransformer
from .transformer import Transformer


from .transformer import Transformer
from ..note import NoteEvent, NoteOnEvent, NoteOffEvent
from ..chord import ChordEvent, ChordOnEvent, ChordOffEvent
from collections import defaultdict
from typing import Callable

class DecomposeChordsTransformer( Transformer ):
    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if isinstance( event, ChordEvent ) or isinstance( event, ChordOnEvent ) or isinstance( event, ChordOffEvent ):
                for note in event.notes:
                    self.add_output( note )
            else:
                self.add_output( event )


from musikla.libraries.keyboard_pynput import KeyboardPynputLibrary
from musikla.libraries.keyboard_mido import KeyboardMidoLibrary
from musikla.core import Context, Library, Music, Value
from musikla.parser import Parser, Node, ParserError, MusiklaExecutionError
from musikla.audio import Player, InteractivePlayer
from musikla.audio.sequencers import FluidSynthSequencerFactory, ABCSequencerFactory, PDFSequencerFactory, HTMLSequencerFactory, MidiSequencerFactory, DebugSequencerFactory
from musikla.libraries import StandardLibrary, MusicLibrary, KeyboardLibrary, MidiLibrary
from typing import Optional, Tuple, Union, Set, Dict, List, Any, cast
from pathlib import Path
from configparser import ConfigParser
import asyncio
import os

def load_config () -> ConfigParser:
    config_path = Path.home() / 'musikla.ini'

    config = ConfigParser()

    if os.path.isfile( config_path ):
        config.read( config_path )
    
    return config

def load_paths ( config : ConfigParser, section : str, option : str ) -> List[str]:
    paths_str = config.get( section, option, fallback = None )

    if paths_str is None:
        return []
    
    return paths_str.split( ":" )

class Script:
    def __init__ ( self, code : Union[str, Node] = None, context : Context = None, config : ConfigParser = None, symbols : Dict[str, Any] = {}, parser_read_cache : bool  = True ):
        self.prelude_context : Context = context or Context.create()
        self.context : Context = self.prelude_context.fork( symbols = self.prelude_context.symbols.fork( True ) )
        self.parser : Parser = Parser( read_cache = parser_read_cache )
        self.player : Player = Player()
        self.config : ConfigParser = config or load_config()
        self.tasks : Set[asyncio.Task] = set()
        self.soundfont : Optional[str] = None
        self.import_paths : List[str] = load_paths( self.config, "Musikla", "path" )
        self.import_cache : Dict[str, Context] = {}
        self.import_extensions : List[str] = [ '.py', '.mkl' ]
        self.files_cache : List[Tuple[str, str]] = []
        self.print_tracebacks : bool = False

        self.player.custom_error_printer = self.print_error

        self.import_paths.insert( 0, Path( __file__ ).parent / 'libraries' )

        self.prelude_context.symbols.assign( 'script', self, local = True )
        # Some core code may depend on the script variable to get a hold of it at runtime
        # But in some child context, the user might have locally binded a custom value to the symbol named "script"
        # And thus any services that were handed that child context would have difficulty finding the main script variable
        # Therefore we publish it too in the "internal" container, which is not available through the musikla language
        # And so we can ensure with more safety that it will be the correct script variable
        self.prelude_context.symbols.assign( 'script', self, local = True, container = 'internal' )

        # This context will be considered main for this execution
        # This allows scripts to detect when they are being imported (and should only declare symbols)
        # versus when they are being executed (and should setup keyboards/play stuff/etc...)
        self.context.symbols.assign( '__main__', True, local = True )

        for key, value in symbols.items():
            self.context.symbols.assign( key, value, local = True )

        self.libraries : Dict[str, Any] = {}
        
        self.add_sequencer_factory( ABCSequencerFactory )
        self.add_sequencer_factory( PDFSequencerFactory )
        self.add_sequencer_factory( HTMLSequencerFactory )
        self.add_sequencer_factory( MidiSequencerFactory )
        self.add_sequencer_factory( DebugSequencerFactory )
        self.add_sequencer_factory( FluidSynthSequencerFactory )

        # Import the builtin libraries
        self.import_library( StandardLibrary, self.player, context = self.prelude_context )
        self.import_library( MusicLibrary, context = self.prelude_context )
        self.import_library( KeyboardLibrary, self.player, context = self.prelude_context )
        self.import_library( KeyboardPynputLibrary, context = self.prelude_context )
        self.import_library( KeyboardMidoLibrary, context = self.prelude_context )
        self.import_library( MidiLibrary, context = self.prelude_context )

        for path in load_paths( self.config, 'Musikla', 'prelude' ):
            resolved_path = self.resolve_import( None, path, None )

            self.import_module( self.prelude_context, resolved_path )

        for path in load_paths( self.config, 'Musikla', 'autoload' ):
            resolved_path = self.resolve_import( None, path, None )

            self.import_module( self.context, resolved_path )

        if code != None:
            self.eval( code )
    
    def add_sequencer_factory ( self, factory : Any ):
        self.player.add_sequencer_factory( factory, self.context, self.config )

    def add_library ( self, *libraries : Any ):
        for library in libraries:
            name : str = library.__name__

            if name.endswith( 'Library' ):
                name = name[ :-7 ]

            self.libraries[ name.lower() ] = library

    def resolve_import ( self, current_path : Optional[str], import_path : str, local : Optional[bool] ) -> str:
        if local is True or local is None:
            if os.path.isabs( import_path ):
                return import_path
            
            if current_path is not None:
                joined_path = str( Path( current_path ).parent.joinpath( import_path ).resolve() )

                if os.path.exists( joined_path ):
                    return joined_path
                
                for ext in self.import_extensions:
                    if os.path.exists( joined_path + ext ):
                        return joined_path + ext

        if local is False or local is None:
            for path in self.import_paths:
                joined_path = str( Path( path ).joinpath( import_path ).resolve() )

                if any( joined_path.endswith( ext ) for ext in self.import_extensions ):
                    if os.path.exists( joined_path ):
                        return joined_path
                
                for ext in self.import_extensions:
                    if os.path.exists( joined_path + ext ):
                        return joined_path + ext

    def import_module ( self, context : Context, module_path : str, save_cache : bool = True ):
        module_path = os.path.realpath( module_path )

        if module_path not in self.import_cache:
            module_context : Context = self.create_subcontext( self.prelude_context, fork = True )

            self.execute_file( module_path, module_context, fork = False, silent = True )

            if save_cache:
                self.import_cache[ module_path ] = module_context
        else:
            module_context : Context = self.import_cache[ module_path ]

        if module_context is not None:
            context.symbols.import_from( module_context.symbols, local = True )

    def import_library ( self, library : Union[str, Library, Any], *args : Any, context : Context = None ):
        context = context or self.context

        if type( library ) is str:
            if library in self.libraries:
                lib_instance = cast( Library, self.libraries[ library ]( *args ) )

                context.link( lib_instance, self )
            else:
                raise Exception( f'Trying to import library {library} not found.' )
        elif isinstance( library, Library ):
            context.link( library, self )
        elif issubclass( library, Library ):
            context.link( library( *args ), self )
        else:
            raise Exception( f'Trying to import library {library} not found.' )

    def parse ( self, code : str, rule : str = None ) -> Node:
        return self.parser.parse( code, rule = rule )

    def play ( self, music : Music, sync : bool = True, realtime = True ) -> Optional[asyncio.Task]:
        if sync:
            self.player.play_more( music )

            self.player.join()
        else:
            # TODO Instead of music, it should eval the thing
            async_player = InteractivePlayer( lambda: list( music ), self.player, realtime = realtime )

            task = asyncio.create_task( async_player.start() )

            self.tasks.add( task )

            task.add_done_callback( lambda a: self.tasks.remove( task ) )

            return task
    
    def create_subcontext ( self, context = None, fork : bool = True, **kargs ) -> Context:
        context = ( context or self.context )
        
        if fork:
            context = context.fork( symbols = context.symbols.fork() )

        for key, value in kargs.items():
            context.symbols.assign( key, value, local = True )

        return context

    def eval ( self, code : Union[str, Node], context : Context = None, fork : bool = False, locals : Dict[str, Any] = {} ) -> Any:
        try:
            if type( code ) is str:
                code = self.parse( code )

            if context is None:
                context = self.create_subcontext( None, fork = fork, **locals )
            elif locals:
                context = self.create_subcontext( context, fork = fork, **locals )

            return Value.eval( context, code )
        except BaseException as err:
            self.print_error( err )
    
    def execute ( self, code : Union[str, Node], context : Context = None, fork : bool = False, silent : bool = False, sync : bool = False, realtime : bool = True ):
        try:
            value = self.eval( code, context = context, fork = fork )

            if value and isinstance( value, Music ):
                if not silent:
                    return self.play( value, sync = sync, realtime = realtime )
                else:
                    # Since code can be lazy, we need to make sure we actually execute the file even
                    # when we have no intention of handling it's events
                    for _ in value.expand( context ): pass

            return None
        except BaseException as err:
            self.print_error( err )
    
    def print_error ( self, err : BaseException ):
        if isinstance( err, ParserError ):
            print( err )
        elif isinstance( err, MusiklaExecutionError ):
            file : Optional[str] = None
            contents : Optional[str] = None

            if err.reporter is None and err.position is not None and err.position[ 0 ] >= 0 and err.position[ 0 ] < len( self.files_cache ):
                file, contents = self.files_cache[ err.position[ 0 ] ]

                err.create_reporter( file, contents )

            print( err )
            if self.print_tracebacks and err.__traceback__ is not None:
                import traceback
                traceback.print_tb(err.__traceback__)
        else:
            print( str( err ) )

    def execute_file ( self, file : str, context : Context = None, fork : bool = False, silent : bool = False, sync : bool = False, realtime : bool = True, locals : Dict[str, Any] = {} ):
        try:
            with open( file, 'r', encoding="utf-8" ) as f:
                file_id = len( self.files_cache )

                file_contents = f.read()

                self.files_cache.append( ( file, file_contents ) )

                code = self.parser.parse( file_contents, file = file, file_id = file_id )
            
            absolute_file : str = os.path.abspath( file )

            value = self.eval( code, context = context, fork = fork, locals = {
                '__file__': absolute_file,
                '__dir__': str( Path( absolute_file ).parent ),
                **locals
            } )
            
            if value and isinstance( value, Music ):
                if not silent:
                    return self.play( value, sync = sync, realtime = realtime )
                else:
                    # Since code can be lazy, we need to make sure we actually execute the file even
                    # when we have no intention of handling it's events
                    for _ in value.expand( context ): pass
        except BaseException as err:
            self.print_error( err )

        return None

    async def join ( self ):
        for task in list( self.tasks ):
            await task

        self.player.sequencers[ 0 ].join()



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .context_modifier_node import ContextModifierNode
from musikla.core.events import ContextChangeEvent
from musikla.core import Context, Voice

class TempoModifierNode( ContextModifierNode ):
    def __init__ ( self, tempo, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.tempo = tempo

    def apply ( self, voice : Voice ):
        voice.tempo = self.tempo

    def modify ( self, context : Context ):
        context.voice = context.voice.clone( tempo = self.tempo )

        yield ContextChangeEvent( context.cursor, "tempo", context.voice.tempo, context.voice, 0 )
        
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'T' + str( self.tempo ) )



from musikla.core import Clock, Scheduler
from musikla.core.events import MusicEvent, VoiceEvent, ControlChangeEvent, ProgramChangeEvent, NoteOnEvent, NoteOffEvent
from musikla.core.events.transformers import DecomposeChordsTransformer, DecomposeNotesTransformer
from .sequencer import Sequencer, SequencerFactory, ArgumentParser
from typing import List, Optional, Tuple
from pathlib import Path
import mido 

class MidiSequencer ( Sequencer ):
    def __init__ ( self, port : str = None, filename : str = None, virtual : bool = False, last_time : int = 0, filter_voices : List[Tuple[str, int]] = [] ):
        super().__init__()

        self.filename : Optional[str] = filename
        self.port : Optional[str] = port
        self.clock : Clock = Clock( auto_start = False )
        self.realtime = self.port is not None
        self.virtual : bool = virtual
        self.filter_voices : List[Tuple[str, int]] = filter_voices

        self.scheduler : Scheduler = Scheduler()
        self.port_obj : Optional[mido.ports.BaseOutput] = None
        self.file_obj : Optional[mido.MidiFile] = None
        self.track_obj : Optional[mido.MidiTrack] = None

        self.last_time : int = last_time

        self.set_transformers(
            DecomposeChordsTransformer(),
            DecomposeNotesTransformer()
        )

    @property
    def playing ( self ):
        return True
        
    def get_time ( self ):
        return self.clock.elapsed()

    def test_event ( self, event : MusicEvent ):
        if not self.filter_voices or not isinstance( event, VoiceEvent ) or event.voice is None:
            return True
        
        for voice, _ in self.filter_voices:
            if voice == event.voice.name:
                return True
            
        return False

    def get_event_channel ( self, event : MusicEvent ):
        if not self.filter_voices or not isinstance( event, VoiceEvent ) or event.voice is None:
            return 0
        
        for voice, channel in self.filter_voices:
            if voice == event.voice.name:
                return channel
            
        return 0

    def _event_to_midi ( self, event : MusicEvent, ticks_per_beat = None ) -> mido.Message:
        if not self.test_event( event ):
            return None

        channel = self.get_event_channel( event )

        if ticks_per_beat is not None:
            t = float( event.timestamp ) / 1000
            lt = float( self.last_time ) / 1000
            delta = mido.second2tick( t, ticks_per_beat ) - mido.second2tick( lt, ticks_per_beat )

            time = delta
        else:
            time = event.timestamp

        self.last_time = event.timestamp

        if isinstance( event, ControlChangeEvent ):
            return mido.Message( 'control_change', channel = channel, control = event.control, value = event.value, time = time )
        elif isinstance( event, ProgramChangeEvent ):
            return mido.Message( 'program_change', channel = channel, program = event.program, time = time )
        elif isinstance( event, NoteOnEvent ):
            return mido.Message( 'note_on', channel = channel, note = int( event ), velocity = event.velocity, time = time )
        elif isinstance( event, NoteOffEvent ):
            return mido.Message( 'note_off', channel = channel, note = int( event ), time = time )

    def on_event ( self, event : MusicEvent ):
        if self.track_obj is not None:
            message = self._event_to_midi( event, self.file_obj.ticks_per_beat )

            if message is None:
                return

            self.track_obj.append( message )

        if self.port_obj is not None:
            message = self._event_to_midi( event )

            if message is None:
                return

            self.scheduler.enqueue( message.time + self.clock.start_time, self.port_obj.send, args = ( message, ) )

    def on_close ( self ):
        if self.file_obj is not None:
            self.file_obj.save( self.filename )
        
        if self.port_obj is not None and not self.port_obj.closed:
            self.port_obj.close()

            self.scheduler.stop()

    def join ( self ):
        pass

    def start ( self ):
        if self.filename is not None:
            self.file_obj = mido.MidiFile()

            self.track_obj = self.file_obj.add_track()
        
        if self.port is not None:
            self.port_obj = mido.open_output( self.port, virtual = self.virtual )
            self.scheduler.start()

        self.clock.start()

class MidiSequencerFactory( SequencerFactory ):
    def init ( self ):
        self.name = 'midi'
        self.argparser = ArgumentParser( description = 'Output events to a MIDI file or as MIDI messages to an output port' )
        self.argparser.add_argument( '-p', '--port', dest = 'port', default = False, action='store_true', help = 'Force this output to be treated as a port' )
        self.argparser.add_argument( '--virtual', dest = 'virtual', default = False, action='store_true', help = 'Whether this is a virtual port to be created' )
        self.argparser.add_argument( '-v', '--voice', dest = 'voices', type = str, action = 'append', help = 'Only accept events from this voices. Can have an optional channel number to route the event to, ex: "piano:2"' )

    def from_str ( self, uri : str, args ) -> Optional[MidiSequencer]:
        suffix = ( Path( uri ).suffix or '' ).lower()

        is_port = args.port or args.virtual

        voices = []
        for voice_name in args.voices or []:
            if ':' in voice_name:
                voice_name, channel = voice_name.split( ':' )
                
                voices.append( ( voice_name, int( channel ) ) )
            else:
                voices.append( ( voice_name, 0 ) )

        if ( suffix == '.midi' or suffix == '.mid' ) and not is_port:
            return MidiSequencer( filename = uri, filter_voices = voices )
        elif uri.startswith( 'midi://' ):
            return MidiSequencer( port = uri[ len( 'midi://' ): ], virtual = args.virtual, filter_voices = voices )
        elif is_port:
            return MidiSequencer( port = uri, virtual = args.virtual, filter_voices = voices )




from musikla.parser.printer import CodePrinter
from typing import Tuple
from .context_modifier_node import ContextModifierNode
from musikla.core.events import ContextChangeEvent
from musikla.core import Context, Voice

class SignatureModifierNode( ContextModifierNode ):
    def __init__ ( self, upper = None, lower = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.upper = upper
        self.lower = lower

    def apply ( self, voice : Voice ):
        if self.upper != None or self.lower != None:
            voice.time_signature = ( self.upper, self.lower )
        elif self.upper != None:
            voice.time_signature = ( self.upper, voice.time_signature[ 1 ] )
        elif self.lower != None:
            voice.time_signature = ( voice.time_signature[ 0 ], self.lower )

    def modify ( self, context : Context ):
        if self.upper != None or self.lower != None:
            context.voice = context.voice.clone( time_signature = ( self.upper, self.lower ) )
        elif self.upper != None:
            context.voice = context.voice.clone( time_signature = ( self.upper, context.time_signature[ 1 ] ) )
        elif self.lower != None:
            context.voice = context.voice.clone( time_signature = ( context.time_signature[ 0 ], self.lower ) )

        yield ContextChangeEvent( context.cursor, "timeSignature", context.voice.time_signature, context.voice, 0 )
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'S' + str( self.upper ) + '/' + str( self.lower ) )



from typing import Tuple
from ..music_node import MusicSequenceBase
from musikla.parser.printer import CodePrinter

class StatementsListNode( MusicSequenceBase ):
    def __init__ ( self, nodes, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.statements = nodes

    def to_source ( self, printer : CodePrinter ):
        for i in range( len( self.statements ) ):
            if i > 0:
                printer.add_token( ';' )
            
            printer.begin_line()

            self.statements[ i ].to_source( printer )

    def values ( self, context ):
        for node in self.statements:
            yield node.eval( context )



from .sequencer import Sequencer, SequencerFactory, ArgumentParser, ArgumentParserError
from .fluidsynth import FluidSynthSequencer, FluidSynthSequencerFactory
from .abc import ABCSequencer, ABCSequencerFactory
from .debug import DebugSequencer, DebugSequencerFactory
from .html import HTMLSequencer, HTMLSequencerFactory
from .pdf import PDFSequencer, PDFSequencerFactory
from .midi import MidiSequencer, MidiSequencerFactory


from typing import Generator, Tuple, AsyncGenerator, Union

class Transformer():
    @staticmethod
    def pipeline2 ( *transformers : 'Transformer' ) -> Tuple['Transformer', 'Transformer']:
        if not transformers:
            raise Exception( "Pipeline cannot be empty" )

        first = transformers[ 0 ]
        latest = first

        for transformer in transformers[ 1: ]:
            latest = latest.pipe_to( transformer )

        return first, latest
    
    @staticmethod
    def pipeline ( *transformers : 'Transformer' ) -> 'Transformer':
        return Transformer.pipeline2( *transformers )[ 0 ]

    @classmethod
    def subscriber ( cls, on_value, on_end = None ):
        transformer = cls()

        transformer.subscribe( on_value, on_end )

        return transformer


    def __init__ ( self ):
        self.subscriptions = []
        self.generator : Generator = self.transform()
        self.input_ended : bool = False
        self.output_ended : bool = False

        self.generator.send( None )

    def transform ( self ):
        while True:
            ended, event = yield

            if ended: break

            self.add_output( event )

    def add_output ( self, item ):
        self._notify_value( item )

    def end_output ( self ):
        if not self.output_ended:
            self.output_ended = True

            self._notify_end()

    def add_input ( self, item ):
        try:
            result = self.generator.send( ( False, item ) )
            
            if result != None:
                self.add_output( result )
        except StopIteration:
            self.end_output()

    def end_input ( self ):
        if not self.input_ended:
            self.input_ended = True

            while not self.output_ended:
                try:
                    result = self.generator.send( ( True, None ) )
                    
                    if result != None:
                        self.add_output( result )
                except StopIteration:
                    self.end_output()

                    break

    def _notify_value ( self, value ):
        for on_value, _ in self.subscriptions:
            if on_value != None: on_value( value )
        
    def _notify_end ( self ):
        for _, on_end in self.subscriptions:
            if on_end != None: on_end()

    def subscribe ( self, on_value, on_end = None ):
        self.subscriptions.append( ( on_value, on_end ) )

    def pipe_to ( self, transformer : 'Transformer', return_self : bool = False ):
        self.subscribe( lambda value: transformer.add_input( value ), lambda: transformer.end_input() )

        if return_self:
            return self
        else:
            return transformer

    @classmethod
    def iter ( cls, it, *args, **kargs ):
        """
        Receives a sync iterable and processes it with this transformer,
        returning the resulting sync iterator
        """
        inst : 'Transformer' = cls( *args, **kargs )

        def generator ():
            nonlocal inst, it

            it = iter( it )

            buffer = []

            inst.subscribe( lambda v: buffer.append( v ) )

            while True:
                for item in buffer: yield item

                buffer.clear()

                try:
                    item = next( it )

                    inst.add_input( item )

                    if inst.output_ended:
                        if hasattr( it, 'close' ) and callable( it.close ):
                            it.close()
                except StopIteration:
                    inst.end_input()

                    break

            for item in buffer: yield item

            buffer.clear()

        return generator()

    @classmethod
    def aiter ( cls, it, *args, **kargs ):
        """
        Receives either a sync or async iterable and processes it with this transformer,
        returning the resulting async iterator
        """
        inst : 'Transformer' = cls( *args, **kargs )

        async def generator ():
            nonlocal inst, it

            if not hasattr( it, '__aiter__' ) and hasattr( it, '__iter__' ):
                it = iter_to_aiter( it )

            it = it.__aiter__()

            buffer = []

            inst.subscribe( lambda v: buffer.append( v ) )

            while True:
                for item in buffer: yield item
                
                buffer.clear()

                try:
                    inst.add_input( await it.__anext__() )
                    
                    if inst.output_ended:
                        if hasattr( it, 'aclose' ) and callable( it.aclose ):
                            await it.aclose()
                except StopAsyncIteration:
                    inst.end_input()
                    break

            for item in buffer: yield item

            buffer.clear()

        return generator()

async def iter_to_aiter ( it ):
    for item in it:
        yield item



from typing import Callable, Any, Optional, cast
from ..voice import Voice
from copy import copy

class MusicEvent():
    def __init__ ( self, timestamp : int = 0, staff : Optional[int] = 0 ):
        self.timestamp : int = timestamp
        self.staff : Optional[int] = staff

    @property
    def end_timestamp ( self ) -> int:
        if hasattr( self, 'duration' ):
            return self.timestamp + cast( DurationEvent, self ).duration
        else:
            return self.timestamp
        
    @end_timestamp.setter 
    def end_timestamp( self, value ):
        if hasattr( self, 'duration' ):
            self.duration = value - self.timestamp
        
            if hasattr( self, 'value' ) and hasattr( self, 'value' ):
                self.value = cast( VoiceEvent, self ).voice.from_duration_absolute( self.duration )


    def clone ( self, **kargs ):
        instance = copy( self )

        for key, value in kargs.items():
            setattr( instance, key, value )
        
        return instance

    def join ( self, context ):
        if context.cursor < self.end_timestamp:
            context.cursor = self.end_timestamp

        return self

    def __add__ ( self, other ):
        return self
        
    def __sub__ ( self, other ):
        return self

    def __repr__ ( self ):
        return "<%s>(%r)" % (self.__class__.__name__, self.__dict__)

class VoiceEvent(MusicEvent):
    def __init__ ( self, timestamp : int = 0, voice : Voice = None, staff : Optional[int] = 0 ):
        super().__init__( timestamp, staff )

        self.voice : Voice = voice or Voice.unknown

class DurationEvent ( VoiceEvent ):
    def __init__ ( self, timestamp = 0, duration = 0, value = 0, voice : Voice = None, staff : Optional[int] = 0 ):
        super().__init__( timestamp, voice, staff )

        # Value stores information about the note duration independent of the tempo and time signature
        self.value = value
        # While duration stores the note's duration as milliseconds
        self.duration = duration

class CallbackEvent ( MusicEvent ):
    def __init__ ( self, timestamp : int, callback : Callable, data : Any = None ):
        super().__init__( timestamp )

        self.callback : Callable = callback
        self.data  : Any= data

    def call ( self ):
        self.callback( self.timestamp, self.data )



class Interval:
    @staticmethod
    def octaves_to_semitones ( octaves : int ) -> int:
        return ( octaves * 12 )

    def __init__ ( self, semitones : int = 0, octaves : int = 0 ):
        self.full_semitones : int = semitones + Interval.octaves_to_semitones( octaves )
    
    @property
    def semitones ( self ) -> int:
        return self.full_semitones % 12

    @property
    def octaves ( self ) -> int:
        return self.full_semitones // 12

    def __int__ ( self ):
        return self.full_semitones

    def __add__ ( self, interval ):
        if interval is None:
            return

        if type( interval ) == int:
            return Interval( self.full_semitones + interval )
        elif isinstance( interval, Interval ):
            return Interval( self.full_semitones + interval.full_semitones )

    def __sub__ ( self, interval ):
        if interval is None:
            return

        if type( interval ) == int:
            return Interval( self.full_semitones - interval )
        elif isinstance( interval, Interval ):
            return Interval( self.full_semitones - interval.full_semitones )


from typing import Hashable, Optional, Set

class Pointer:
    def __init__ ( self, scope, name ):
        self.scope = scope
        self.name = name

    def __str__ ( self ):
        return f"Pointer@{ id( self.scope ) }\{ self.name }"

"""
A note about the terminology here:
    - Opaque symbol scopes are scopes that, when assigned to, will always default to storing the value in themselves.
      This means that if a parent scope has a symbol with the same name, then it will be shadowed by the new one.
      If the intention is to use the parent's symbol, then a pointer needs to be created first (with the `using` method)
    - Transparent symbol scopes, on the other hand, before assigning a new symbol, will perform a shallow search up their
      scope tree for a symbol with the same name, and if found, create a pointer automatically.
    - Shallow searches are recursive searches that stop at the first opaque symbol scope they find.
"""

class ValueNotFound:
    pass

class SymbolsScope:
    def __init__ ( self, parent = None, opaque : bool = True ):
        self.parent : SymbolsScope = parent
        self.symbols : dict = dict()
        self.opaque : bool = opaque

    @property
    def root ( self ) -> 'SymbolsScope':
        scope : SymbolsScope = self

        while scope.parent is not None:
            scope = scope.parent
        
        return scope

    @property
    def prelude ( self ) -> Optional['SymbolsScope']:
        scope : SymbolsScope = self

        while scope is not None:
            if scope.lookup( "is_prelude", 'internal', recursive = False ):
                return scope
            
            scope = scope.parent
        
        return None

    def pointer ( self, name : Hashable, container : str = "", shallow : bool = False ) -> Pointer:
        scope : SymbolsScope = self

        while scope != None:
            value = scope.lookup( name, container = container, recursive = False, follow_pointers = False, default = ValueNotFound )

            if value != ValueNotFound:
                if isinstance( value, Pointer ):
                    scope = value.scope
                    name = value.name

                break

            if shallow and scope.opaque:
                scope = None
            else:
                scope = scope.parent

        if scope == None:
            return None
        else:
            return Pointer( scope, name )

    def using ( self, name : Hashable, alias : str = None, container : str = "", shallow : bool = False, soft : bool = False ) -> 'SymbolsScope':
        if isinstance( name, Pointer ):
            pointer = name
            name = pointer.name
        elif name != None and self.parent != None:
            pointer = self.parent.pointer( name, container = container, shallow = shallow )
        else:
            pointer = None
        
        if pointer == None and not soft:
            raise BaseException( f"Trying to use global undefined symbol { name }" )
        elif pointer != None:
            self.assign( alias or name, pointer, container = container, local = True )
        
        return pointer.scope if pointer != None else None

    def enumerate ( self, prefix : str = None, limit : int = None, ignore : Set = None, container : str = "", local : bool = False ):
        if ignore is None:
            ignore = set()

        count = 0

        if container in self.symbols:
            for key, value in self.symbols[ container ].items():
                if limit is not None:
                    if count >= limit:
                        return
                    else:
                        count += 1
                
                if key in ignore:
                    continue

                if prefix is not None and not key.startswith( prefix ):
                    continue

                ignore.add( key )

                yield key, value

        
        if not local and self.parent is not None:
            if limit is not None:
                limit -= count

            for key, value in self.parent.enumerate( prefix = prefix, limit = limit, ignore = ignore, container = container ):
                yield key, value


    def lookup ( self, name : Hashable, container : str = "", recursive : bool = True, follow_pointers : bool = True, default = None, raise_default : bool = False, stop_on : Optional['SymbolsScope'] = None ):
        if stop_on is None or self != stop_on:
            if container in self.symbols and name in self.symbols[ container ]:
                value = self.symbols[ container ][ name ]

                if follow_pointers and isinstance( value, Pointer ):
                    return value.scope.lookup( value.name, container = container )

                return value
            
            if recursive and self.parent != None:
                return self.parent.lookup( name, container = container, follow_pointers = follow_pointers, default = default, raise_default = raise_default, stop_on = stop_on )
        
        if raise_default:
            raise default
        else:
            return default

    def assign ( self, name : Hashable, value, container = "", follow_pointers : bool = True, local : bool = True ):
        if container not in self.symbols:
            self.symbols[ container ] = dict()

        if name not in self.symbols[ container ]:
            if not local and not self.opaque:
                self.using( name, container = container, shallow = True, soft = True )

        if follow_pointers:
            existing_value = self.lookup( name, container = container, recursive = False, follow_pointers = False )

            if existing_value != None and isinstance( existing_value, Pointer ):
                existing_value.scope.assign( existing_value.name, value, container = container )

                return

        self.symbols[ container ][ name ] = value

    def lookup_instrument ( self, name ):
        return self.lookup( name, container = "instruments" )

    def assign_instrument ( self, instrument ):
        self.assign( instrument.name, instrument, container = "instruments" )
        
        return instrument

    def lookup_internal ( self, name ):
        return self.lookup( name, container = "internal" )

    def assign_internal ( self, name, value ):
        self.assign( name, value, container = "internal" )

    def fork ( self, opaque : bool = True ):
        return SymbolsScope( self, opaque )

    def import_from ( self, scope : 'SymbolsScope', local : bool = True ):
        for key, value in scope.enumerate( local = local ):
            # Treat symbols that start with an underscore as private
            if not key.startswith( "_" ):
                self.assign( key, value, local = True )

    def unref ( self ):
        for name, value in self.symbols:
            if callable( getattr( value, 'unref', None ) ):
                value.unref()

class Ref:
    def __init__ ( self, pointer : Pointer, container : str = '' ):
        self.pointer : Pointer = pointer
        self.container : str = container

    def get ( self ):
        return self.pointer.scope.lookup( self.pointer.name, container = self.container )

    def set ( self, value ):
        self.pointer.scope.assign( self.pointer.name, value, container = self.container )



from typing import Optional
from musikla.core.events.transformers import Transformer, ComposeNotesTransformer, ComposeChordsTransformer, VoiceIdentifierTransformer, AnnotateTransformer, EnsureOrderTransformer
from musikla.core.events import MusicEvent, NoteEvent, ProgramChangeEvent
from musikla.core import Clock
from ..sequencer import Sequencer, SequencerFactory
from .builder import ABCBuilder
from pathlib import Path

class ABCSequencer ( Sequencer ):
    def __init__ ( self, filename : str ):
        super().__init__()

        self.file_builder : ABCBuilder = ABCBuilder()
        self.filename : str = filename
        self.clock : Clock = Clock( auto_start = False )

        self.set_transformers(
            # EnsureOrderTransformer( 'beforeCompose' ),
            ComposeNotesTransformer(),
            ComposeChordsTransformer(),
            # EnsureOrderTransformer( 'afterCompose' ),
            VoiceIdentifierTransformer(),
            # EnsureOrderTransformer( 'afterIdentify', False ),
            AnnotateTransformer()
        )
    
    @property
    def playing ( self ) -> bool:
        return False
        
    def get_time ( self ):
        if not self.clock.started:
            return 0

        return self.clock.elapsed()

    def on_event ( self, event : MusicEvent ):
        self.file_builder.add_event( event )

    def on_close ( self ):
        if self.filename is not None:
            with open( self.filename, 'w' ) as f:    
                file = self.file_builder.build()

                f.write( str( file ) )
                
                f.flush()

    def join ( self ):
        pass

    def start ( self ):
        self.clock.start()

class ABCSequencerFactory( SequencerFactory ):
    def init ( self ):
        self.name = 'abc'

    def from_str ( self, uri : str, args ) -> Optional[ABCSequencer]:
        suffix = ( Path( uri ).suffix or '' ).lower()

        if suffix == '.abc':
            return ABCSequencer( uri )


from musikla.audio.sequencers.abc.sequencer import ABCSequencer
from musikla.core.events import MusicEvent
from .sequencer import Sequencer, SequencerFactory
from typing import Optional
from pathlib import Path
from tempfile import gettempdir, gettempprefix

class PDFSequencer ( Sequencer ):
    def __init__ ( self, filename : str ):
        super().__init__()

        self.filename : str = filename

        temp : str = f"{ gettempdir() }/{ gettempprefix() }-{ Path( filename ).name }.abc"

        self.abc_sequencer : ABCSequencer = ABCSequencer( temp )
    
    @property
    def playing ( self ):
        return self.abc_sequencer.playing
        
    def get_time ( self ):
        return self.abc_sequencer.get_time()

    def on_event ( self, event : MusicEvent ):
        self.abc_sequencer.register_event( event )
    
    def on_close ( self ):
        self.abc_sequencer.close()

    def join ( self ):
        pass

    def start ( self ):
        self.abc_sequencer.start()

class PDFSequencerFactory( SequencerFactory ):
    def init ( self ):
        self.name = 'pdf'

    def from_str ( self, uri : str, args ) -> Optional[PDFSequencer]:
        suffix = ( Path( uri ).suffix or '' ).lower()

        if suffix == '.pdf':
            return PDFSequencer( uri )
        



from typing import Any, Optional, Union, cast
from ..voice import Voice
from .event import DurationEvent
from wave import Wave_read, open
from pathlib import Path
from subprocess import Popen, PIPE
from weakref import WeakValueDictionary
import os

class SoundEvent( DurationEvent ):
    cached_sounds : WeakValueDictionary = WeakValueDictionary()

    @staticmethod
    def is_optimized ( file : str ) -> bool:
        suffix : str = Path( file ).suffix.lower()

        if suffix == '.wav' or suffix == '.wave':
            wave : Wave_read = open( file, 'rb' )

            return wave.getnchannels() == 1 and wave.getsampwidth() == 2 and wave.getframerate() == 44100
        else:
            return False

    @staticmethod
    def optimize_folder ( input : str, output : str ):
        for file in os.listdir( input ):
            if os.path.isfile( os.path.join( input, file ) ):
                SoundEvent.optimize( os.path.join( input, file ), os.path.join( output, file ) )

    @staticmethod
    def optimize ( input : str, output : str ):
        p = Popen( 
            [ "ffmpeg", '-y', '-hide_banner', "-loglevel", "panic", "-i", input, "-vn", "-c:a", "pcm_s16le", "-ac", "1", '-ar', '44100', "-f", "wav", output ]
        )

        p.wait()

    @staticmethod
    def convert ( file : str ) -> Wave_read:
        p = Popen( 
            [ "ffmpeg", '-y', '-hide_banner', "-loglevel", "panic", "-i", file, "-vn", "-c:a", "pcm_s16le", "-ac", "1", '-ar', '44100', "-f", "wav", "-" ], 
            stdout = PIPE
        )

        return open( cast( Any, p.stdout ), 'rb' )
    
    @staticmethod
    def open_wave ( file : str ) -> Wave_read:
        suffix : str = Path( file ).suffix.lower()

        if suffix == '.wav' or suffix == '.wave':
            wave : Wave_read = open( file, 'rb' )

            if wave.getnchannels() != 1 or wave.getsampwidth() != 2 or wave.getframerate() != 44100:
                return SoundEvent.convert( file )

            return wave
        else:
            return SoundEvent.convert( file )
        
    @staticmethod
    def open_wave_cached ( file : str ):
        try:
            return SoundEvent.cached_sounds[ file ]
        except:
            wave = SoundEvent.open_wave( file )

            SoundEvent.cached_sounds[ file ] = wave

            return wave

    def __init__ ( self, file : Union[str, Wave_read], timestamp = 0, duration = None, value = None, velocity : int = 127, voice : Voice = None ):
        super().__init__( timestamp, duration, value, voice )

        self.file : Optional[str] = file if type( file ) is str else None
        self.velocity : int = velocity
        self.wave : Wave_read = SoundEvent.open_wave_cached( file ) if type( file ) is str else file
        self.wave_duration : int = int( ( self.wave.getnframes() / float( self.wave.getframerate() ) ) * 1000 )
        
        if self.duration is None and self.value is None:
            self.duration = self.wave_duration

        if self.duration is not None and self.value is None:
            self.value = self.voice.from_duration_absolute( self.duration )
        elif self.duration is None and self.value is not None:
            self.duration = self.voice.get_duration( float( self.value ) )

    def __str__ ( self ) -> str:
        res = f"sample({ self.file })"

        if self.value != 1:
            res += str( self.value )
        
        return res



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .expression_node import ExpressionNode
from musikla.core import Music

class VariableExpressionNode( ExpressionNode ):
    def __init__ ( self, name, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.name = name

    def assign ( self, context, value, local : bool = False ):
        context.symbols.assign( self.name, value, local = local )

    def lookup_assign ( self, context, value_fn, local : bool = False ):
        value = context.symbols.lookup( self.name, recursive = not local )

        value = value_fn( value )

        context.symbols.assign( self.name, value, local = local )

    def __eval__ ( self, context ):
        value = context.symbols.lookup( self.name )
        
        if isinstance( value, Music ):
            return Music( value.expand( context ) )

        return value

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( '$' + self.name )


from typing import List, Optional, Tuple
from colorama import Fore, Back, Style

class ErrorReporter:
    def __init__ ( self, error_phase : str, error_message : str, contents : str, position : Tuple[int, int], file : str = None ):
        self.error_phase : str = error_phase
        self.error_message : str = error_message
        self.contents : str = contents
        self.position : Tuple[int, int] = position
        self.file : Optional[str] = file

    def _is_newline ( self, char : str ) -> bool:
        return char == '\n'

    # Returns the (start_line, error_line, line_offsets)
    def _get_line_offsets ( self, adjacent_lines : int = 3 ) -> Tuple[int, int, List[int]]:
        error_line : int = -1

        start_line : int = 0

        line_offsets : List[int] = [ 0 ]

        line_offsets_count : int = 1

        line_offsets_index : int = 1

        max_lines : int = adjacent_lines * 2 + 2

        remaining_lines : int = adjacent_lines + 1

        for i in range( len( self.contents ) ):
            if i == self.position[ 0 ]:
                error_line = start_line + line_offsets_count - 1
            
            if self._is_newline( self.contents[ i ] ):
                if line_offsets_count < max_lines:
                    line_offsets.append( i + 1 )

                    line_offsets_count += 1
                else:
                    if error_line >= 0 and start_line == error_line:
                        break

                    line_offsets[ line_offsets_index ] = i + 1

                    start_line += 1
                
                line_offsets_index = ( line_offsets_index + 1 ) % max_lines
                
                if i >= self.position[ 1 ]:
                    remaining_lines -= 1
                
                if remaining_lines == 0:
                    break
        
        if error_line < 0:
            error_line = start_line + line_offsets_count - 1

        return ( start_line, error_line, line_offsets[ line_offsets_index: ] + line_offsets[ :line_offsets_index ] )

    def _number_digits ( self, n ):
        return len( str( int( n ) ) )

    def __str__ ( self ):
        lines : List[str] = []

        def println ( ln = '' ):
            lines.append( ln )

        start_line, error_line, line_offsets = self._get_line_offsets()

        max_digits = self._number_digits( start_line + len( line_offsets ) + 1 )

        println()

        line = str( error_line + 1 )
        col = str( self.position[ 0 ] - line_offsets[ error_line - start_line ] + 1 )
        println( Back.RED + " " + self.error_phase + " " + Style.RESET_ALL + " in " + Fore.CYAN + ( self.file or "<in memory>" ) + Style.RESET_ALL + ":" + line + ":" + col  )
        
        for i in range( len( line_offsets ) - 1 ):
            start = line_offsets[ i ]
            end = line_offsets[ i + 1 ]

            number = start_line + i + 1

            number_str = format( number, '0' + str(max_digits) ) + " |"

            if start_line + i == error_line:
                number_str = Fore.RED + number_str + Style.RESET_ALL

            if self.position[ 0 ] >= start and self.position[ 1 ] <= end:
                line_str = self.contents[ start:self.position[ 0 ] ] \
                         + Back.RED + self.contents[ self.position[ 0 ]:self.position[ 1 ] ] + Style.RESET_ALL \
                         + self.contents[ self.position[ 1 ]:end - 1 ]
            elif self.position[ 0 ] >= start and self.position[ 0 ] < end and self.position[ 1 ] >= end:
                line_str = self.contents[ start:self.position[ 0 ] ] \
                         + Back.RED + self.contents[ self.position[ 0 ]:end - 1 ] + Style.RESET_ALL
            elif self.position[ 0 ] < start and self.position[ 1 ] > start and self.position[ 1 ] <= end:
                line_str = Back.RED + self.contents[ start:self.position[ 1 ] ] + Style.RESET_ALL \
                         + self.contents[ self.position[ 1 ]:end - 1 ]
            elif self.position[ 0 ] < start and self.position[ 1 ] > end:
                line_str = Back.RED + self.contents[ start:end - 1 ] + Style.RESET_ALL
            else:
                line_str = self.contents[ start:end ]

            println( f" {number_str}  {line_str}".strip() )
        
        println()
        println( Fore.RED + "Message " + Style.RESET_ALL + self.error_message )
        println()

        return '\n'.join( lines )
    
    def print ( self ):
        print( str( self ) )



from musikla.core.events import MusicEvent, VoiceEvent, ContextChangeEvent, NoteEvent, ChordEvent, RestEvent, BarNotationEvent, StaffNotationEvent
from musikla.core.theory import NotePitchClassesInv
from fractions import Fraction
from .file import ABCFile, ABCStaff, ABCVoice, ABCBar, ABCNote, ABCChord, ABCRest, ABCVoiceGroup, ABCVoiceGroupKind
from typing import Dict, Optional
import re

class ABCBuilderConfig:
    def __init__ ( self, visible_voice_names : bool = True, score_merge_voice_staffs : bool = False ):
        self.visible_voice_names : bool = visible_voice_names
        self.score_merge_voice_staffs : bool = score_merge_voice_staffs

class ABCBuilder:
    def __init__ ( self, config : ABCBuilderConfig = None ):
        self.file : ABCFile = ABCFile()

        self.config : ABCBuilderConfig = config or ABCBuilderConfig()

    def get_current_staff ( self, voice : str ) -> ABCStaff:
        if voice not in self.file.body.staffs:
            self.file.body.staffs[ voice ] = []
        
        if len( self.file.body.staffs[ voice ] ) == 0:
            staff = ABCStaff()

            self.file.body.staffs[ voice ].append( staff )
        else:
            staff = self.file.body.staffs[ voice ][ -1 ]

        return staff

    def get_current_bar ( self, voice : str ) -> ABCBar:
        staff = self.get_current_staff( voice )

        if len( staff.bars ) == 0:
            bar = ABCBar()

            staff.bars.append( bar )
        else:
            bar = staff.bars[ -1 ]

        return bar

    def add_context_change ( self, event : ContextChangeEvent ):
        if event.property == 'length':
            self.file.header.length = Fraction( event.value )
        elif event.property == 'timeSignature':
            self.file.header.meter = event.value
        elif event.property == 'tempo':
            self.file.header.tempo = event.value

    def add_voice_to_score ( self, voice : ABCVoice ):
        match = re.match( r"^(.+?)_[0-9]+$", voice.name_escaped )
        existing_voice : Optional[ABCVoice] = None

        if match is not None:
            existing_voice = next( v for v in self.file.header.voices if v.name_escaped.startswith( match[ 1 ] ) )

        if existing_voice is not None:
            self.file.header.score.groups.append_with( voice.name_escaped, existing_voice.name_escaped )
        else:
            group = ABCVoiceGroup( ABCVoiceGroupKind.Parentheses, [ voice.name_escaped ] )

            self.file.header.score.groups.append( group )

    def add_voice ( self, event : VoiceEvent ):
        if self.file.header.length is None:
            self.file.header.length = Fraction( event.voice.value )
        
        if self.file.header.meter is None:
            self.file.header.meter = event.voice.time_signature

        if self.file.header.tempo is None:
            self.file.header.tempo = event.voice.tempo

        if event.voice.name not in self.file.body.staffs:
            self.file.body.staffs[ event.voice.name ] = []

            voice = ABCVoice()
            voice.name = event.voice.name

            if self.config.visible_voice_names:
                voice.label = event.voice.name

            self.file.header.voices.append( voice )
            
            if self.config.score_merge_voice_staffs:
                self.add_voice_to_score( voice )

    def build_note ( self, event : NoteEvent, inside_chord : bool = False ) -> ABCNote:
        note = ABCNote()

        note.accidental = event.accidental
        note.octave = event.octave
        
        if not inside_chord:
            note_length = Fraction( event.value ) / Fraction( self.file.header.length )

            note.length = note_length

        note.pitch_class = NotePitchClassesInv[ event.pitch_class ]
        note.tied = not inside_chord and event.tied

        return note

    def add_note ( self, event : NoteEvent ):
        current_bar = self.get_current_bar( event.voice.name )

        note = self.build_note( event )

        current_bar.symbols.append( note )

    def add_chord ( self, event : ChordEvent ):
        chord_length = Fraction( event.value ) / Fraction( self.file.header.length )

        current_bar = self.get_current_bar( event.voice.name )

        chord = ABCChord()

        chord.notes = [ self.build_note( n, True ) for n in event.notes ]
        chord.length = chord_length
        chord.name = event.name
        chord.tied = event.tied

        current_bar.symbols.append( chord )

    def add_rest ( self, event : RestEvent ):
        rest_length = Fraction( event.value ) / Fraction( self.file.header.length )

        current_bar = self.get_current_bar( event.voice.name )

        rest = ABCRest()

        rest.length = rest_length
        rest.visible = event.visible

        current_bar.symbols.append( rest )

    def add_bar ( self, event : BarNotationEvent ):
        staff = self.get_current_staff( event.voice.name )

        staff.bars.append( ABCBar() )

    def add_staff ( self, event : StaffNotationEvent ):
        name = event.voice.name
        if name not in self.file.body.staffs:
            self.file.body.staffs[ name ] = [ ABCStaff() ]
        else:
            self.file.body.staffs[ name ].append( ABCStaff() )

    def add_event ( self, event : MusicEvent ):
        if isinstance( event, ContextChangeEvent ):
            self.add_context_change( event )
        else:
            if isinstance( event, VoiceEvent ):
                self.add_voice( event )

            if isinstance( event, NoteEvent ):
                self.add_note( event )
            elif isinstance( event, ChordEvent ):
                self.add_chord( event )
            elif isinstance( event, RestEvent ):
                self.add_rest( event )
            elif isinstance( event, BarNotationEvent ):
                self.add_bar( event )
            elif isinstance( event, StaffNotationEvent ):
                self.add_staff( event )

    def build ( self ) -> ABCFile:
        return self.file



from typing import List, Any, Union
from musikla.core import Context, Music, Voice, Library, Clock
from musikla.core.callable_python_value import CallablePythonValue
from musikla.core.events import MusicEvent, NoteOnEvent, NoteOffEvent, ControlChangeEvent, ProgramChangeEvent
from musikla.core.theory import Note
from musikla.core.events.transformers import ComposeNotesTransformer, TerminationMelodyTransformer, TeeTransformer, VoiceIdentifierTransformer, Transformer
from musikla.libraries.music import MusiklaNotationBuilderTransformer
import mido
from mido.ports import MultiPort

def midi_to_event ( time : int, msg : Any, voice : Voice ):
    if msg.type == 'control_change':
        return ControlChangeEvent( time, voice, msg.control, msg.value )
    elif msg.type == 'program_change':
        pass
        # return ProgramChangeEvent( time, voice, msg.program )
    elif msg.type == 'note_on':
        note = Note().with_pitch( msg.note )

        return NoteOnEvent( 
            timestamp = time, 
            pitch_class = note.pitch_class, 
            octave = note.octave, 
            accidental = note.accidental, 
            velocity = msg.velocity, 
            voice = voice 
        )
    elif msg.type == 'note_off':
        note = Note().with_pitch( msg.note )

        return NoteOffEvent( 
            timestamp = time, 
            pitch_class = note.pitch_class, 
            octave = note.octave, 
            accidental = note.accidental,
            voice = voice 
        )
        
    return None

def midi_track_to_music ( 
        context, mid, stream : Any, voice : Voice, 
        ignore_message_types : List[str] = None 
    ):
    # TODO Read tempo from midi file
    tempo = 416666

    ticks = 0

    for msg in stream:
        ticks += msg.time

        if ignore_message_types is not None and msg.type in ignore_message_types:
            continue

        time = context.cursor + int( mido.tick2second( ticks, mid.ticks_per_beat, tempo ) * 1000 )

        event = midi_to_event( time, msg, voice )

        if event:
            yield event
        else:
            pass

def midi_stream_to_music ( 
        context, stream : Any, voice : Voice, 
        ignore_message_types : List[str] = None 
    ):
    clock = Clock()

    start = context.cursor

    for msg in stream:
        if ignore_message_types is not None and msg.type in ignore_message_types:
            continue

        time = start + clock.elapsed()

        event = midi_to_event( time, msg, voice )

        if event:
            yield event
        else:
            pass

def read_midi_file (
        context : Context, 
        file : mido.MidiFile = None,
        voices : List[Voice] = None,
        cutoff_sequence = None,
        ignore_message_types : List[str] = None
    ):
    tracks = []

    for i, track in enumerate( file.tracks ):
        if voices is None:
            msgVoice = context.voice
        else:
            msgVoice = voices[ i ]

        if msgVoice == None:
            continue

        events = midi_track_to_music( context, file, track, msgVoice, ignore_message_types )
            
        events = ComposeNotesTransformer.iter( events )

        tracks.append( events )

    return Music.parallel( tracks )

def function_readmidi (
        context : Context, 
        file : str = None, 
        port : Union[str, List[str], bool] = None, 
        voices : List[Voice] = None,
        cutoff_sequence = None,
        ignore_message_types : List[str] = None
    ):
    if file is not None:
        mid = mido.MidiFile( file )

        return read_midi_file( context, mid, voices, cutoff_sequence, ignore_message_types )
    else:
        if port == True:
            default_port = context.library( MidiLibrary ).get_midi_default_input()

            if default_port is None:
                port = mido.open_input()
            else:
                port = mido.open_input( default_port )
        elif type(port) == str:
            port = mido.open_input( port )
        elif type(port) == list:
            port = MultiPort( [ mido.open_input( p ) for p in port ] )
        
        events = midi_stream_to_music( context, port, context.voice, ignore_message_types )

        events = ComposeNotesTransformer.iter( events )

        if cutoff_sequence != None:
            events = TerminationMelodyTransformer.iter( events, list( cutoff_sequence.expand( context.fork( cursor = 0 ) ) ) )

        events = TeeTransformer.iter( 
            events, 
            Transformer.pipeline(
                VoiceIdentifierTransformer(),
                MusiklaNotationBuilderTransformer(),
                Transformer.subscriber( lambda n: print( n + '\n\n' ) )
            )
        )

        return Music( list( events ) )

def function_get_midi_input_name ( context : Context ) -> str:
    return context.library( MidiLibrary ).get_midi_default_input()

def function_set_midi_input_name ( context : Context, name : str ):
    context.library( MidiLibrary ).set_midi_default_input( name )

class MidiLibrary(Library):
    def on_link ( self, script ):
        context : Context = self.context

        context.symbols.assign( "readmidi", CallablePythonValue( function_readmidi ) )
        context.symbols.assign( "get_midi_input_name", CallablePythonValue( function_get_midi_input_name ) )
        context.symbols.assign( "set_midi_input_name", CallablePythonValue( function_set_midi_input_name ) )

    def set_midi_default_input ( self, name : str ):
        self.context.symbols.assign( "midi_default_input", name, container = "internal" )

    def get_midi_default_input ( self ) -> str:
        return self.context.symbols.lookup( "midi_default_input", container = "internal" )



from .event import MusicEvent, DurationEvent, VoiceEvent
from .note import NoteEvent, NoteOnEvent, NoteOffEvent
from ..voice import Voice
from ..theory import Note, NoteAccidental, Interval
from fractions import Fraction
from typing import Dict, List, Optional

class ChordEvent( DurationEvent ):
    def __init__ ( self, timestamp = 0, pitches : List[int] = [], name : str = None, duration = 4, voice : Voice = None, velocity = 127, value = None, tied : bool = False, staff : Optional[int] = 0 ):
        super().__init__( timestamp, duration, value, voice, staff )

        self.name : Optional[str] = name
        self.pitches : List[int] = pitches
        self.velocity = velocity
        self.tied : bool = tied

    @property
    def notes ( self ) -> List[NoteEvent]:
        return [ self.note_at( i ) for i in range( len( self.pitches ) ) ]

    def note_at ( self, index : int ) -> NoteEvent:
        return NoteEvent.from_pitch( 
            timestamp = self.timestamp,
            pitch = self.pitches[ index ],
            duration = self.duration,
            voice = self.voice,
            velocity = self.velocity,
            value = self.value,
            tied = self.tied,
            staff = self.staff
        )

    def music ( self ):
        from ..music import SharedMusic

        return SharedMusic( [ self ] )

    @property
    def chord_on ( self ) -> 'ChordOnEvent':
        return ChordOnEvent( 
            timestamp = self.timestamp, 
            pitches = self.pitches, 
            name = self.name, 
            voice = self.voice, 
            velocity = self.velocity, 
            tied = self.tied,
            staff = self.staff
        )

    @property
    def chord_off ( self ) -> 'ChordOffEvent':
        return ChordOffEvent( 
            timestamp = self.timestamp + self.duration, 
            pitches = self.pitches, 
            name = self.name, 
            voice = self.voice, 
            tied = self.tied,
            staff = self.staff
        )

    def with_root_pitch ( self, pitch : int, **kargs ) -> 'ChordEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        rp : int = self.pitches[ 0 ] if self.pitches else 0

        return ChordEvent(
            timestamp = self.timestamp,
            pitches = [ pitch + ( p - rp ) for p in self.pitches ],
            name = name,
            duration = self.duration, 
            voice = self.voice, 
            velocity = self.velocity, 
            value = self.value,
            staff = self.staff
        )

    def with_pitches ( self, pitches : List[int], **kargs ) -> 'ChordEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        return ChordEvent(
            timestamp = self.timestamp,
            pitches = pitches,
            name = name,
            duration = self.duration, 
            voice = self.voice, 
            velocity = self.velocity, 
            value = self.value,
            staff = self.staff
        )

    def __eq__ ( self, other ):
        if other is None or not isinstance( other, ChordEvent ):
            return False

        return len( self.pitches ) == len( other.pitches ) \
           and all( p1 == p2 for p1, p2 in zip( self.pitches, other.pitches ) )

    def __add__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p + interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p + int( interval ) for p in self.pitches ] )
        else:
            return self

    def __sub__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p - interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p - int( interval ) for p in self.pitches ] )
        else:
            return self

    def __repr__ ( self ):
        return f'[{self.timestamp}]' + str( self )

    def __str__ ( self ):
        notes = ''.join( str( n ) for n in self.notes )

        if self.name is None:
            return f"[{ notes }]"

        return f'"{ self.name }"[{ notes }]'

class ChordOnEvent( VoiceEvent ):
    def __init__ ( self, timestamp = 0, pitches : List[int] = [], name : str = None, voice : Voice = None, velocity = 127, tied : bool = False, staff : Optional[int] = 0 ):
        super().__init__( timestamp, voice, staff )

        self.name : Optional[str] = name
        self.pitches : List[int] = pitches
        self.velocity = velocity
        self.tied : bool = tied

    def chord_off ( self, timestamp : int ) -> 'ChordOffEvent':
        return ChordOffEvent( 
            timestamp = timestamp, 
            pitches = self.pitches, 
            name = self.name, 
            voice = self.voice, 
            tied = self.tied,
            staff = self.staff
        )

    @property
    def notes ( self ) -> List[NoteOnEvent]:
        return [ self.note_at( i ) for i in range( len( self.pitches ) ) ]

    def note_at ( self, index : int ) -> NoteOnEvent:
        return NoteOnEvent.from_pitch( 
            timestamp = self.timestamp,
            pitch = self.pitches[ index ],
            voice = self.voice,
            velocity = self.velocity,
            tied = self.tied,
            staff = self.staff
        )

    def music ( self ):
        from ..music import SharedMusic

        return SharedMusic( [ self ] )

    def with_root_pitch ( self, pitch : int, **kargs ) -> 'ChordOnEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        rp : int = self.pitches[ 0 ] if self.pitches else 0

        return ChordOnEvent(
            timestamp = self.timestamp,
            pitches = [ pitch + ( p - rp ) for p in self.pitches ],
            name = name,
            voice = self.voice, 
            velocity = self.velocity,
            staff = self.staff
        )

    def with_pitches ( self, pitches : List[int], **kargs ) -> 'ChordOnEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        return ChordOnEvent(
            timestamp = self.timestamp,
            pitches = pitches,
            name = name,
            voice = self.voice, 
            velocity = self.velocity, 
            staff = self.staff
        )

    def __eq__ ( self, other ):
        if other is None or not isinstance( other, ChordEvent ):
            return False

        return len( self.pitches ) == len( other.pitches ) \
           and all( p1 == p2 for p1, p2 in zip( self.pitches, other.pitches ) )

    def __add__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p + interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p + int( interval ) for p in self.pitches ] )
        else:
            return self

    def __sub__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p - interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p - int( interval ) for p in self.pitches ] )
        else:
            return self

    def __repr__ ( self ):
        return f'[{self.timestamp}]' + str( self )

    def __str__ ( self ):
        notes = ''.join( str( n.note ) for n in self.notes )

        if self.name is None:
            return f"[{ notes }](On)"

        return f'"{ self.name }"[{ notes }](On)'

class ChordOffEvent( VoiceEvent ):
    def __init__ ( self, timestamp = 0, pitches : List[int] = [], name : str = None, voice : Voice = None, tied : bool = False, staff : Optional[int] = 0 ):
        super().__init__( timestamp, voice, staff )

        self.name : Optional[str] = name
        self.pitches : List[int] = pitches
        self.tied : bool = tied

    @property
    def notes ( self ) -> List[NoteOffEvent]:
        return [ self.note_at( i ) for i in range( len( self.pitches ) ) ]

    def note_at ( self, index : int ) -> NoteOffEvent:
        return NoteOffEvent.from_pitch( 
            timestamp = self.timestamp,
            pitch = self.pitches[ index ],
            voice = self.voice,
            tied = self.tied,
            staff = self.staff
        )

    def music ( self ):
        from ..music import SharedMusic

        return SharedMusic( [ self ] )

    def with_root_pitch ( self, pitch : int, **kargs ) -> 'ChordOffEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        rp : int = self.pitches[ 0 ] if self.pitches else 0

        return ChordOffEvent(
            timestamp = self.timestamp,
            pitches = [ pitch + ( p - rp ) for p in self.pitches ],
            name = name,
            voice = self.voice,
            staff = self.staff
        )

    def with_pitches ( self, pitches : List[int], **kargs ) -> 'ChordOffEvent':
        name = kargs[ 'name' ] if 'name' in kargs else self.name

        return ChordOffEvent(
            timestamp = self.timestamp,
            pitches = pitches,
            name = name,
            voice = self.voice,
            staff = self.staff
        )

    def __eq__ ( self, other ):
        if other is None or not isinstance( other, ChordEvent ):
            return False

        return len( self.pitches ) == len( other.pitches ) \
           and all( p1 == p2 for p1, p2 in zip( self.pitches, other.pitches ) )

    def __add__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p + interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p + int( interval ) for p in self.pitches ] )
        else:
            return self

    def __sub__ ( self, interval ):
        if type( interval ) == int:
            return self.with_pitches( [ p - interval for p in self.pitches ] )
        elif isinstance( interval, Interval ):
            return self.with_pitches( [ p - int( interval ) for p in self.pitches ] )
        else:
            return self

    def __repr__ ( self ):
        return f'[{self.timestamp}]' + str( self )

    def __str__ ( self ):
        notes = ''.join( str( n.note ) for n in self.notes )

        if self.name is None:
            return f"[{ notes }](Off)"

        return f'"{ self.name }"[{ notes }](Off)'


from musikla.core.enumerable import IterCursor, merge_sorted
from typing import Optional
from musikla.audio.interactive_player import InteractivePlayer
from musikla.core.context import Context
from musikla.core.events.event import MusicEvent
from typing import List, Dict, Tuple, Iterable, Iterator, Generator, TypeVar
from musikla.libraries.keyboard.action import KeyAction
from musikla.core.music import Music
from musikla.libraries.keyboard.keyboard import Keyboard
from musikla.libraries.keyboard.event import KeyboardEvent
from musikla.core.events.transformers import Transformer, ComposeNotesTransformer

T = TypeVar( 'T' )

def concat_iter ( cursor : IterCursor[T], iterable : Iterable[T] ) -> Iterable[T]:
    while not cursor.ended:
        yield cursor.current

        cursor.move_next()
    
    for n in iterable:
        yield n

class VirtualPlayerThread:
    def __init__ ( self, player : InteractivePlayer, iterable : Iterable[MusicEvent] ):
        self.player : InteractivePlayer = player
        self.iterable : Iterable[MusicEvent] = iterable
        self.iterator : Optional[IterCursor[MusicEvent]] = None
        
    def concat ( self, iterable : Iterable[MusicEvent] ):
        if self.iterator is None:
            self.iterator = IterCursor( self.iterable )

            self.iterator.move_next()
        
        self.iterable = concat_iter( self.iterator, iterable )
        self.iterator = None

    def seek ( self, time : int ) -> Iterable[MusicEvent]:
        if self.iterator is None:
            self.iterator = IterCursor( self.iterable )

            self.iterator.move_next()

        while not self.iterator.ended:
            if self.iterator.current is not None:
                if self.iterator.current.timestamp <= time or time < 0:
                    yield self.iterator.current
                else:
                    break

            self.iterator.move_next()

class VirtualPlayer:
    def __init__ ( self ):
        self.queued_interactions : List[Tuple[int, str, KeyboardEvent]] = []
        self._active_actions : Dict[KeyAction, Tuple[KeyAction, Optional[InteractivePlayer]]] = {}
        self._output : List[MusicEvent] = []

        self._transformers : Transformer = ComposeNotesTransformer()
        self._transformers.subscribe( lambda ev: self._output.append( ev ) )

        self._active_key : Optional[KeyAction] = None
        self._active_player : Optional[InteractivePlayer] = None

        self._player_threads : List[VirtualPlayerThread] = []

        self.time : int = 0

    def get_time ( self ) -> int:
        return self.time

    def play_more ( self, events : Iterable[MusicEvent] ):
        if self._active_key is None or self._active_key.interactive_player is None:
            raise Exception( "Trying to play sequence with no key or player attached" )

        if self._active_key.interactive_player == self._active_player:
            for thread in self._player_threads:
                if thread.player == self._active_player:
                    thread.concat( events )
                    break
        else:
            self._player_threads.append( VirtualPlayerThread( self._active_key.interactive_player, events ) )

    def on_press ( self, time : int, key : KeyboardEvent ):
        self.queued_interactions.append( ( time, 'press', key ) )

    def on_release ( self, time : int, key : KeyboardEvent ):
        self.queued_interactions.append( ( time, 'release', key ) )

    def _seek ( self, until : int ):
        iterators = [ thread.seek( until ) for thread in self._player_threads ]

        for ev in merge_sorted( iterators, lambda ev: ev.timestamp ):
            self.time = ev

            self._transformers.add_input( ev )

        if until >= 0:
            self.time = until
    
    def _perform ( self, context : Context, kind : str, key : KeyboardEvent, keyboards : List[Keyboard] ):
        actions : List[KeyAction] = []

        for kb in keyboards:
            if key in kb:
                actions.append( kb[ key ] )
        
        for action in actions:
            virtual_action : KeyAction

            if action in self._active_actions:
                virtual_action, _ = self._active_actions[ action ]
            else:
                virtual_action = action.clone( 
                    sync = True,
                    is_pressed = False,
                    interactive_player = None 
                )

                self._active_actions[ action ] = ( virtual_action, None )

            self._active_key = virtual_action
            self._active_player = virtual_action.interactive_player

            if kind == 'press':
                virtual_action.on_press( context, self, key.get_parameters() )
            elif kind == 'release':
                virtual_action.on_release( context, self )

            self._active_actions[ action ] = ( virtual_action, virtual_action.interactive_player )

            self._active_key = None
            self._active_player = None
        
    def _generate_music ( self, context : Context, keyboards : List[Keyboard] ):
        # 1. Loop while queue not empty
        for i in range( len( self.queued_interactions ) ):
            # 1.1. Fetch next interaction
            time, kind, key = self.queued_interactions[ i ]

            # 1.3. Seek until next interaction
            self._seek( time )

            # 1.2. Update active actions list
            self._perform( context.fork( cursor = time ), kind, key, keyboards )

        # 2. Seek until the end
        self._seek( -1 )

        # 3. Collect and sort generated events in parallel
        self._transformers.end_input()

        return self._output

    def generate ( self, context : Context, keyboards : List[Keyboard] ) -> Music:
        return Music( self._generate_music( context, keyboards ) )


from musikla.parser.printer import CodePrinter
from typing import Optional, Tuple
from ..node import Node
from musikla.core import Context

class MacroNode(Node):
    def __init__ ( self, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.virtual_node : Optional[Node] = None

    def __eval__ ( self, context : Context ):
        return self.virtual_node.eval( context )

    def to_source ( self, printer : CodePrinter ):
        self.virtual_node.to_source( printer )


from .macro import MacroNode
from .keyboard import MacroNode, KeyboardShortcutMacroNode, KeyboardDeclarationMacroNode, KeyboardShortcutDynamicMacroNode, KeyboardShortcutComprehensionMacroNode, KeyboardForLoopMacroNode, KeyboardWhileLoopMacroNode, KeyboardIfMacroNode, KeyboardBlockMacroNode
from .voice import VoiceDeclarationMacroNode



from typing import Tuple, List, Dict, Optional, Union
from musikla.core import Context, Library
from musikla.core.callable_python_value import CallablePythonValue
from musikla.core import Music
from musikla.core.events import NoteEvent, SoundEvent, ChordEvent
from musikla.core.events.transformers import Transformer, VoiceIdentifierTransformer, NotationBuilderTransformer, DecomposeChordsTransformer
from musikla.parser.abstract_syntax_tree import Node
from musikla.parser.abstract_syntax_tree.music_parallel_node import Box
from musikla.parser.abstract_syntax_tree.macros import VoiceDeclarationMacroNode

## NotationBuilder ##
from musikla.core.voice import Voice
from musikla.core.events.event import MusicEvent
from musikla.core.theory import Note, Chord
from musikla.parser.printer import CodePrinter
import musikla.core.events as evt
import musikla.parser.abstract_syntax_tree as ast
import musikla.parser.abstract_syntax_tree.expressions as ast_exp
import musikla.parser.abstract_syntax_tree.statements as ast_stm
import musikla.parser.abstract_syntax_tree.context_modifiers as ast_mod

class MusiklaNotationBuilderTransformer(NotationBuilderTransformer):
    def __init__ ( self, 
        only_final : bool = False, 
        ast : bool = False, 
        context : Context = None, 
        nameless_voices : bool = True, 
        declare_voices : bool = True, 
        base_voice : Voice = None 
    ):
        """
        only_final          When true, the result is only emitted after the entire 
                            input (musical sequence) has been processed. When false,
                            intermediate notations are continuously emitted.
        
        ast                 When true, emits the Abstract Syntax Tree objects. When
                            false, transforms those objects into a string that can
                            be parsed, and emits the string instead

        nameless_voices     When true, it tries to produce notation that carries with
                            it any state that is needed, instead of prepending 
                            the names of the voices the events belonged to
        
        declare_voices      When `nameless_voices = False`, also include the 
                            definitions of the named voices used.
        
        base_voice          ...
        """
        super().__init__( only_final )

        self.ast : bool = ast
        self.context : Context = context or Context.create()
        self.nameless_voices : bool = nameless_voices
        self.declare_voices : bool = declare_voices
        self.base_voice : Voice = base_voice or Voice()

    def _get_voice_name ( self, voice : Voice ) -> str:
        return voice.name.split( "/" )[ 0 ]


    def note_to_ast ( self, context : Context, voice : Voice, event : NoteEvent ) -> ast.NoteNode:
        return ast.NoteNode( Note(
            pitch_class = event.pitch_class,
            octave = event.octave - voice.octave,
            accidental = event.accidental,
            value = voice.get_relative_value( event.value )
        ) )

    def event_to_ast ( self, context : Context, voice : Voice, events : List[ast.Node], event : MusicEvent ):
        if not any( isinstance( event, eventType ) for eventType in [ evt.NoteEvent, evt.ChordEvent, evt.RestEvent ] ):
            return

        if isinstance( event, evt.NoteEvent ) and event.velocity != voice.velocity:
            voice = voice.clone( velocity = event.velocity )

            events.append( ast_mod.VelocityModifierNode( event.velocity ) )
        
        if isinstance( event, evt.NoteEvent ):
            events.append( self.note_to_ast( context, voice, event ) )
        elif isinstance( event, evt.ChordEvent ):
            events.append( ast.ChordNode( Chord(
                notes = [ self.note_to_ast( context, voice, n ).note for n in event.notes ],
                name = event.name,
                value = voice.get_relative_value( event.value )
            ) ) )
        elif isinstance( event, evt.RestEvent ):
            events.append( ast.RestNode(
                value = voice.get_relative_value( event.value ),
                visible = event.visible
            ) )

    def event_sequence_to_ast ( self, context : Context, voice: Voice, events : List[MusicEvent] ) -> ast.MusicSequenceNode:
        nodes : List[ast.Node] = []

        for ev in events:
            self.event_to_ast( context, voice, nodes, ev )

        return ast.MusicSequenceNode( nodes )

    def modifiers_to_ast ( self, base_voice : Voice, actual_voice : Voice, music : ast.MusicSequenceNode = None ) -> Optional[ast.MusicSequenceNode]:
        """
        Compares the settings between `base_voice` and `actual_voice`. When a
        difference is found, creates a AST node describing the value on 
        `actual_voice`.
        """
        mods : List[ast_mod.ContextModifierNode] = []

        if base_voice.time_signature != actual_voice.time_signature: # (4, 4):
            u, d = actual_voice.time_signature
            mods.append( ast_mod.SignatureModifierNode( u, d ) )

        if base_voice.tempo != actual_voice.tempo: # 60:
            mods.append( ast_mod.TempoModifierNode( actual_voice.tempo ) )

        if base_voice.octave != actual_voice.octave: # 4:
            mods.append( ast_mod.OctaveModifierNode( actual_voice.octave ) )

        if base_voice.value != actual_voice.value: # 1:
            mods.append( ast_mod.LengthModifierNode( actual_voice.value ) )

        if base_voice.velocity != actual_voice.velocity: # 127:
            mods.append( ast_mod.VelocityModifierNode( actual_voice.velocity ) )

        if mods and music:
            mods.extend( music.expressions )

            return ast.MusicSequenceNode( mods )
        elif mods:
            return ast.MusicSequenceNode( mods )
        else:
            return music or None

    def voice_to_ast ( self, context: Context, voice : Voice, music : ast.MusicSequenceNode = None, base_voice : Voice = None ) -> Union[ast_mod.VoiceBlockModifier, ast.MusicSequenceNode, None]:
        if self.nameless_voices:
            return self.modifiers_to_ast( base_voice or self.base_voice, voice, music = music )
        else:
            name = self._get_voice_name( voice )
            
            music = self.modifiers_to_ast( base_voice or voice, voice, music = music )

            if name == context.voice.name:
                return music
            else:
                return ast_mod.VoiceBlockModifier( music, name )
        
    def voice_sequence_to_ast ( self, context : Context, voice : Tuple[Voice, int, List[MusicEvent]], base_voice : Voice = None ) -> Union[ast_mod.VoiceBlockModifier, ast.MusicSequenceNode, None]:
        sequence = self.event_sequence_to_ast( context, voice[ 0 ], voice[ 2 ] )

        if not sequence.expressions:
            return None
        
        return self.voice_to_ast( context, voice[ 0 ], sequence, base_voice = base_voice )

    def voice_declaration_to_ast ( self, base_voice : Voice, actual_voice : Voice ) -> VoiceDeclarationMacroNode:
        name = self._get_voice_name( actual_voice )

        return VoiceDeclarationMacroNode(
            name = name,
            modifiers = self.modifiers_to_ast( base_voice, actual_voice ) or ast_exp.MusicSequenceNode()
        )

    def to_ast ( self, events_per_voice : List[Tuple[Voice, int, List[MusicEvent]]] ) -> ast.Node:
        l = len( events_per_voice )

        if l == 0:
            return ast.MusicSequenceNode( [] )
        elif l == 1:
            staff = self.voice_sequence_to_ast( self.context.fork(), events_per_voice[ 0 ] )

            # When there are no notes on this staff, there's also no need to even
            # generate the voice declarations. We just return an empty group ()
            if staff is None:
                return ast_exp.GroupNode()
            
            # decl will contain the voice declaration used for this staff, if
            # voice declarations are enabled
            decl : Optional[VoiceDeclarationMacroNode] = None

            if not self.nameless_voices and self.declare_voices:
                decl = self.voice_declaration_to_ast( self.base_voice, events_per_voice[ 0 ][ 0 ] )
    
                return ast_stm.StatementsListNode( [ decl, staff ] )

            return staff
        else:
            decls : List[VoiceDeclarationMacroNode] = []
            # Sometimes voices can have the same name but slightly different
            # settings. When that happens, we want to keep a reference to what
            # version of that voice we created a declaration for. That way we can
            # generate any extras modifiers needed to apply those changes
            decls_voices : Dict[str, Voice] = {}

            staffs : List[ast.Node] = []

            for voice_staff in events_per_voice:
                voice, _, _ = voice_staff
                
                name = self._get_voice_name( voice )

                base_voice = self.base_voice

                if not self.nameless_voices and self.declare_voices:
                    if name not in decls_voices:
                        decls.append( self.voice_declaration_to_ast( self.base_voice, voice ) )

                        decls_voices[ name ] = voice

                        base_voice = voice

                staff_expr = self.voice_sequence_to_ast( self.context.fork(), voice_staff, base_voice = base_voice )

                if staff_expr is not None:
                    staffs.append( staff_expr )

            if not staffs:
                return ast_exp.GroupNode()

            music_expr = ast.MusicParallelNode( staffs )

            if not decls:
                return music_expr
            
            return ast_stm.StatementsListNode( [
                *decls,
                music_expr
            ] )

    def to_string ( self, events_per_voice : List[Tuple[Voice, int, List[MusicEvent]]] ) -> Union[str, ast.Node]:
        if self.ast:
            return self.to_ast( events_per_voice )
        else:
            ast = self.to_ast( events_per_voice )

            printer = CodePrinter()

            return printer.print( ast )

def function_sample ( context : Context, file : str, duration : float = None, len = None ):
    event = SoundEvent( file, timestamp = context.cursor, voice = context.voice, duration = duration, value = len, velocity = context.voice.velocity )

    context.cursor += event.duration

    return Music( [ event ] )

def function_transpose ( note, semitones : int = 0, octaves : int = 1 ):
    if isinstance( note, NoteEvent ):
        note = note.clone(
            octave = note.octave + octaves
        )
    
    return note

def function_to_mkl ( context : Context, music : Music, 
        ast : bool = False, 
        nameless_voices: bool = True, 
        declare_voices : bool = True, 
        base_voice : Voice = None  ) -> Union[str, Node]:
    source : Box[str] = Box( "" )

    inp, out = Transformer.pipeline2(
        VoiceIdentifierTransformer(), 
        MusiklaNotationBuilderTransformer( 
            only_final = True, 
            ast = ast, 
            context = context, 
            nameless_voices = nameless_voices,
            declare_voices = declare_voices,
            base_voice = base_voice
        ),
    )

    out.subscribe( lambda s: source.set( s ) )

    for ev in music.expand( context.fork( cursor = 0 ) ):
        inp.add_input( ev )
    
    inp.end_input()

    return source.value

def function_chord ( context : Context, *notes ):
    notes = [ n.first_note() if isinstance( n, Music ) else n for n in notes ]

    chord = ChordEvent(
        timestamp = context.cursor, 
        pitches = [ int(n) for n in notes  ], 
        name = None, 
        duration = notes[ 0 ].duration, 
        voice = notes[ 0 ].voice, 
        velocity = notes[ 0 ].velocity, 
        value = notes[ 0 ].value, 
        tied = notes[ 0 ].tied, 
        staff = notes[ 0 ].staff
    )

    return Music( [ chord ] )

class MusicLibrary(Library):
    def on_link ( self, script ):
        context : Context = self.context

        context.symbols.assign( 'chord', CallablePythonValue( function_chord ) );

        context.symbols.assign( 'is_sample_optimized', CallablePythonValue( SoundEvent.is_optimized ) )
        context.symbols.assign( 'optimize_sample', CallablePythonValue( SoundEvent.optimize ) )
        context.symbols.assign( 'optimize_samples_folder', CallablePythonValue( SoundEvent.optimize_folder ) )
        context.symbols.assign( "sample", CallablePythonValue( function_sample ) )

        context.symbols.assign( "transpose", CallablePythonValue( function_transpose ) )
        context.symbols.assign( "to_mkl", CallablePythonValue( function_to_mkl ) )





from .event import MusicEvent, VoiceEvent, DurationEvent, CallbackEvent
from .note import NoteEvent, NoteOnEvent, NoteOffEvent
from .rest import RestEvent
from .change import ProgramChangeEvent, ControlChangeEvent, ContextChangeEvent
from .chord import ChordEvent, ChordOnEvent, ChordOffEvent
from .notation import BarNotationEvent, StaffNotationEvent, BAR_STANDARD, BAR_DOUBLE, BAR_END, BAR_BEGIN_REPEAT, BAR_END_REPEAT, BAR_BEGIN_END_REPEAT
from .sound import SoundEvent


from musikla.parser.printer import CodePrinter
from typing import Any, Optional, Tuple, Union
from .statement_node import StatementNode
from ..expressions.variable_expression_node import VariableExpressionNode
from ..node import Node
from musikla.core import Value, Context

class VariableDeclarationStatementNode( StatementNode ):
    def __init__ ( self, left : Union[Node, str], right : Node, operator : Optional[str] = None, local : bool = False, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        if type( left ) is str:
            self.left : Any = VariableExpressionNode( left )
        else:
            self.left : Any = left

        self.right : Any = right
        self.operator : Optional[str] = operator
        self.local : bool = local

        if self.operator is None and not hasattr( self.left, 'assign' ):
            raise BaseException( f"Left hand side \"{CodePrinter().print(self.left)}\" cannot be used in an attribution" )
        elif self.operator is not None and not hasattr( self.left, 'lookup_assign' ):
            raise BaseException( f"Left hand side \"{CodePrinter().print(self.left)}\" cannot be used in an attribution" )

    def __eval__ ( self, context : Context ):
        val = Value.assignment( self.right.eval( context.fork( cursor = 0 ) ) )

        if self.operator is None or self.operator == "":
            self.left.assign( context, val, local = self.local )
        else:
            def _set ( value ):
                nonlocal val

                # value = context.symbols.lookup( self.name, recursive = not self.local )
                if self.operator == '*': value *= val
                elif self.operator == '/': value /= val
                elif self.operator == '+': value += val
                elif self.operator == '-': value -= val
                elif self.operator == '&': value &= val
                elif self.operator == '|': value |= val
                else: raise Exception( "Invalid operator: " + self.operator )

                return value

            self.left.lookup_assign( context, _set, local = self.local )

        return None

    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( f" { self.operator if self.operator is not None else '' }= " )
        
        self.right.to_source( printer )



from time import sleep
from typing import Optional
import bimpy
import asyncio
import time
import threading
from musikla.core import Context

class SingleWindow:
    def __init__ ( self, context : Context, title : str, render_fn, width : int = 600, height : int = 600, margin : int = 0, max_framerate : int = 60 ):
        self.context : Context = context
        self.title = title
        self.render_fn = render_fn
        self.width = width
        self.height = height
        self.margin = margin
        self.running : bool = False
        self.max_framerate : int = max_framerate
        self._opened = bimpy.Bool()
        self._bimpy_ctx = None

        self.start()

    def _render ( self, event : threading.Event = None ):
        ctx = self._bimpy_ctx

        with ctx:
            bimpy.set_next_window_pos(bimpy.Vec2(self.margin, self.margin), bimpy.Condition.Always)
            bimpy.set_next_window_size(bimpy.Vec2(ctx.width() - self.margin * 2, ctx.height() - self.margin * 2), bimpy.Condition.Always)

            if bimpy.begin(self.title, self._opened, bimpy.WindowFlags.NoTitleBar | bimpy.WindowFlags.NoResize):
                try:
                    self.render_fn()
                except BaseException as err:
                    self.context.script.print_error(err)

                bimpy.end()
            
            # We set the event so that the event loop is free before we call Render
            if event is not None:
                event.set()


    def _render_async ( self, loop : asyncio.AbstractEventLoop ):
        self._render()

        if not self._bimpy_ctx.should_close():
            loop.call_later(1 / self.max_framerate, self._render_async, loop)
        else:
            self.running = False
        
            self._bimpy_ctx = None
            self._opened = None

    def _block_thread ( self, event_start : threading.Event, event_stop : threading.Event ):
        event_start.set()

        event_stop.wait()
        event_stop.clear()

    def start(self):
        def _start_thread ( loop : asyncio.AbstractEventLoop ):
            event_start = threading.Event()
            event_stop = threading.Event()

            while not ctx.should_close():
                sleep(1.0 / self.max_framerate)
                loop.call_soon_threadsafe(self._block_thread, event_start, event_stop)
                event_start.wait()
                event_start.clear()

                self._render( event_stop )

            self.running = False
        
            self._bimpy_ctx = None
            self._opened = None

            # self._render_async( asyncio.get_event_loop() )
            
        self.running = True
        
        ctx = bimpy.Context()

        ctx.init(self.width, self.height, self.title)

        with ctx:
            bimpy.themes.set_light_theme()

        self._bimpy_ctx = ctx
        self._opened = bimpy.Bool()

        self._render_async( asyncio.get_event_loop() )

        # threading.Thread( target = _start, args = ( asyncio.get_event_loop(), ) ).start()
        
        



import bimpy
from musikla.core.events import MusicEvent, NoteEvent
from musikla.core.events.transformers import DecomposeChordsTransformer
from musikla.core import Voice
from typing import Any, List, Optional, Tuple

class EventsPreview:
    class Viewport:
        def __init__ ( self, width : int, height : int, scroll_vertical : int = 0, scroll_horizontal : int = 0, time_signature = (4, 4), tempo : int = 200, bar_width : int = 60, events_height : float = 2.0 ):
            self.width = width
            self.height = height
            self.time_signature = time_signature
            self.tempo = tempo
            self.bar_width = 200 # bar_width
            self.events_height = events_height
            self.scroll_vertical = scroll_vertical
            self.scroll_horizontal = scroll_horizontal

            ## CACHE
            # Caching these properties allows avoiding recalculating them every
            # frame

            # The cache key is used to know when to invalidate all the calculated values
            self.cache_key = None
            # The voice used to calculate timings inside the viewport
            self.cache_voice : Voice = Voice( "", None, 
                time_signature = self.time_signature,
                tempo = self.tempo
            )
            # How many pixels in width does one second occupy
            self.cache_second_width : int = 0
            # What are the lower and upper pitches that should be visible
            self.cache_fit_pitch_amount : int = 0
            self.cache_lower_pitch : int = 0
            self.cache_upper_pitch : int = 0
            # If the viewport height and event«s height don't subdivide perfectly
            # There can be paddings equally distributed in the top and bottom
            self.cache_vertical_padding : float = 0

        def invalidate_cache ( self ):
            self.cache_key = None

        def validate_cache( self ) -> bool:
            """Checks if the cache is valid. If it is invalid, updates it"""

            key = (self.width, self.height, self.scroll_horizontal, self.scroll_vertical, self.time_signature, self.tempo, self.bar_width, self.events_height)

            if self.cache_key is None or self.cache_key != key:
                self.cache_key = key

                self.cache_voice.time_signature = self.time_signature
                self.cache_voice.tempo = self.tempo

                self.cache_second_width = self.bar_width / ( self.cache_voice.get_bar_duration_absolute() / 1000 )
                
                self.cache_fit_pitch_amount = self.height // self.events_height
                self.cache_vertical_padding = ( self.height % self.events_height ) / 2

                self.cache_lower_pitch = 68 - ( self.cache_fit_pitch_amount // 2 )
                self.cache_upper_pitch = self.cache_lower_pitch + self.cache_fit_pitch_amount

                return False

            return True
        
        def get_pitch_y ( self, pitch : int ) -> Optional[float]:
            if pitch >= self.cache_lower_pitch and pitch <= self.cache_upper_pitch:
                y = self.cache_vertical_padding + ( ( pitch - self.cache_lower_pitch ) * self.events_height )

                return self.height - y

            return None
        
        def get_time_x ( self, time : int, absolute : bool = False ) -> Optional[float]:
            x = ( time * self.cache_second_width / 1000 )

            if absolute:
                return x
            
            x -= self.scroll_horizontal
            
            if x >= 0 and x < self.width:
                return x

            return None

        def scroll_to_time ( self, time : int, anchor : str = 'middle' ):
            time_x = self.get_time_x( time, absolute = True )

            if anchor == 'start':
                self.scroll_horizontal = max( time_x, 0 )
            elif anchor == 'middle':
                self.scroll_horizontal = max( time_x - ( self.width // 2 ), 0 )
            elif anchor == 'end':
                self.scroll_horizontal = max( time_x - self.width, 0 )

    class CachedEvent:
        def __init__ ( self, viewport, event : NoteEvent ):
            self.event : NoteEvent = event

            self.visible : bool = False
            self.start_pos : bimpy.Vec2 = None
            self.end_pos : bimpy.Vec2 = None

            y = viewport.get_pitch_y( int( event ) )

            if y is not None:
                start_x = viewport.get_time_x( event.timestamp )
                end_x = viewport.get_time_x( event.end_timestamp )


                if start_x is not None or end_x is not None:
                    if start_x is None: start_x = 0
                    if end_x is None: end_x = viewport.width

                    self.visible = True

                    self.start_pos = bimpy.Vec2( start_x, y )
                    self.end_pos = bimpy.Vec2( end_x, y )
            

    def __init__ ( self, events : List[MusicEvent], width : int, height : int, viewport = None, margins : Tuple[int, int, int, int] = None ):
        self.events : List[MusicEvent] = events
        # top, right, bottom, left
        self.margins : Tuple[int, int, int, int] = margins or (5, 10, 10, 10)
        self._height : int = height
        self._width : int = width

        self.viewport : EventsPreview.Viewport = viewport or EventsPreview.Viewport( 
            width = width - self.margins_h,
            height = height - self.margins_v,
            bar_width = 60,
            events_height = 2.0
        )

        self.cache_events : Optional[List[EventsPreview.CachedEvent]] = None
        self.cache_rect_start : bimpy.Vec2 = None
        self.cache_rect_end : bimpy.Vec2 = None

    def invalidate_cache ( self ):
        self.cache_events = None

    def validate_cache ( self ) -> bool:
        if self.cache_events is None or not self.viewport.validate_cache():
            self.cache_events = []

            for ev in self.events:
                if isinstance( ev, NoteEvent ):
                    cached_event = EventsPreview.CachedEvent( self.viewport, ev )

                    if cached_event.visible:
                        self.cache_events.append( cached_event )
            
            self.cache_rect_start = bimpy.Vec2( self.margins[ 3 ], self.margins[ 0 ] )
            self.cache_rect_end = bimpy.Vec2( self.width - self.margins[ 1 ], self.height - self.margins[ 2 ] )

            return False
        
        return True

    @property
    def width ( self ) -> int:
        return self._width

    @width.setter
    def width ( self, value ):
        self._width = value
        self.viewport.width = value - self.margins_h

    @property
    def height ( self ) -> int:
        return self._height

    @height.setter
    def height ( self, value ):
        self._height = value
        self.viewport.height = value - self.margins_v

    @property
    def margins_v ( self ) -> int:
        return self.margins[ 0 ] + self.margins[ 2 ]

    @property
    def margins_h ( self ) -> int:
        return self.margins[ 1 ] + self.margins[ 3 ]

    def render ( self ):
        self.validate_cache()
        
        cursor = bimpy.get_cursor_screen_pos()

        rect_start = cursor + self.cache_rect_start
        rect_end = cursor + self.cache_rect_end

        bimpy.add_rect_filled( rect_start, rect_end, 0xFFFFFFFF, 5.0 )

        for ev in self.cache_events:
            if ev.visible:
                bimpy.add_line( rect_start + ev.start_pos, rect_start + ev.end_pos, 0xFF000000, self.viewport.events_height )

        bimpy.set_cursor_pos_y( bimpy.get_cursor_pos_y() + self.height )

class KeyboardBufferPreview:
    def __init__ ( self, buffer, *args, **kargs ):
        from musikla.libraries.keyboard.buffer import KeyboardBuffer

        self.buffer : KeyboardBuffer = buffer
        self.preview : EventsPreview = EventsPreview( [], *args, **kargs )

        self.cache_key = None
        self.cache_events : List[MusicEvent] = []
    
    def invalidate_cache ( self ):
        self.cache_key = None

    def validate_cache ( self ) -> bool:
        len_collected = len( self.buffer.collected_events )
        len_buffered = len( self.buffer.music_buffer.buffer ) \
            if self.buffer.music_buffer is not None else 0

        now_buffer = self.buffer.duration_live

        key = ( len_collected, len_buffered, now_buffer )

        if self.buffer.recording or self.cache_key is None or self.cache_key != key:
            self.cache_key = key

            events = self.buffer.to_events_live( now = True );

            events = DecomposeChordsTransformer.iter( events )

            self.cache_events = list( events )

            self.preview.viewport.scroll_to_time( now_buffer, anchor = 'end' )

            return False
        
        return True

    def render ( self ):
        if not self.validate_cache():
            self.preview.events = self.cache_events
            self.preview.invalidate_cache()
        
        self.preview.render()

        


from musikla.audio.sequencers.abc.sequencer import ABCSequencer
from musikla.core.events import MusicEvent
from .sequencer import Sequencer, SequencerFactory
from typing import Optional
from pathlib import Path
from tempfile import gettempdir, gettempprefix

class HTMLSequencer ( Sequencer ):
    def __init__ ( self, filename : str ):
        super().__init__()

        self.filename : str = filename

        temp : str = f"{ gettempdir() }/{ gettempprefix() }-{ Path( filename ).name }.abc"

        self.abc_sequencer : ABCSequencer = ABCSequencer( temp )
    
    @property
    def playing ( self ):
        return self.abc_sequencer.playing

    def build ( self ) -> str:
        abc = str( self.abc_sequencer.file_builder.build() )

        template = f"""<html>
    <head>
        <link rel="stylesheet" href="http://dev.music.free.fr/css/music.min.css" />
        <script src="http://dev.music.free.fr/js/abc-ui-1.0.0.min.js"></script>
    </head>
    <body>
        <div class="abc-source">
{abc}
        </div>
        <script>
            $ABC_UI.init();
        </script>
    </body>
</html>
"""

        return template

    def get_time ( self ):
        return self.abc_sequencer.get_time()

    def on_event ( self, event : MusicEvent ):
        self.abc_sequencer.register_event( event )
    
    def on_close ( self ):
        self.abc_sequencer.close()

        with open( self.filename, 'w' ) as f:    
            f.write( self.build() )
                
            f.flush()

    def join ( self ):
        pass

    def start ( self ):
        self.abc_sequencer.start()

class HTMLSequencerFactory( SequencerFactory ):
    def init ( self ):
        self.name = 'html'
    
    def from_str ( self, uri : str, args ) -> Optional[HTMLSequencer]:
        suffix = ( Path( uri ).suffix or '' ).lower()

        if suffix == '.html' or suffix == '.htm':
            return HTMLSequencer( uri )
        



from .action import KeyAction
from .event import KeyboardEvent, PianoKey, KeyEvent, KeyStroke, KeyStrokePress, KeyStrokeRelease, MouseClick, MouseMove, MouseScroll, EventSource
from .grid import Grid
from .keyboard import Keyboard
from .virtual_player import VirtualPlayer
from .library import KeyboardLibrary
from .buffer import KeyboardBuffer


from musikla.parser.printer import CodePrinter
from typing import Optional, Tuple
from .node import Node
from musikla.core import Value, StackFrame, Context

class StackFrameNode( Node ):
    def __init__ ( self, child : Optional[Node], position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.child : Optional[Node] = child
    
    def to_source ( self, printer : CodePrinter ):
        if self.child is not None:
            self.child.to_source( printer )

    def __eval__ ( self, context : Context ):
        if self.child is not None:
            context.symbols.assign( 'stack_frame', StackFrame(), container = 'stack' )

            return Value.eval( context, self.child )



from musikla.libraries.keyboard.event import KeyEvent, KeyStrokePress, KeyStrokeRelease
from musikla.core import Context, Library
from musikla.libraries.keyboard import KeyboardLibrary, EventSource, MouseMove, MouseClick, MouseScroll, KeyStroke
from pynput.keyboard import Key
from pynput import keyboard, mouse
from typing import Any, Dict, List, Optional, Tuple, cast
import asyncio

class KeyboardPynputLibrary( Library ):
    def __init__ ( self ):
        super().__init__( "keyboard_pynput" )

    def on_link ( self, script ):
        keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        if keyboard is not None:
            keyboard.add_source( KeyboardPynputEventSource( self.context ) )

class KeyboardPynputEventSource( EventSource ):
    def __init__ ( self, context : Context ):
        self.context : Context = context
        self.keyboard_listener : Optional[keyboard.Listener] = None
        self.mouse_listener : Optional[mouse.Listener] = None
        self.keyboard_state = dict()
        self.keymaps : Dict[int, str] = dict()

        if self.context.script.config.has_section( 'Keyboard.KeyMap' ):
            for value, key in self.context.script.config[ 'Keyboard.KeyMap' ].items():
                self.keymaps[ int( value ) ] = key

    def get_key_info ( self, key : Key ) -> Tuple[bool, str, int]:
        if hasattr( key, '_value_' ):
            value : int = int( key._value_.vk )
        elif hasattr( key, '_scan' ):
            value : int = key._scan
        else:
            value : int = -1

        key_str = str( key )

        # Hardcoded fix for wrong value for 5 numpad key
        if value in self.keymaps: key_str = self.keymaps[ value ]
        else:
            if key_str == '<65437>': key_str = "'5'"
            elif key_str == "'\\\\'": key_str = "'\\'"
            elif key_str == '<220>': key_str = "'\\'"
            elif key_str == '<65032>': key_str = "Key.shift"

            key_str = key_str[ len( 'Key.' ): ] if key_str.startswith( 'Key.' ) else key_str[ 1:-1 ]

        if key_str == 'ctrl_l' or key_str == 'ctrl_r':
            key_str = 'ctrl'
        elif key_str == 'shift_l' or key_str == 'shift_r':
            key_str = 'shift'

        is_modifier : bool = key_str in [ 'ctrl', 'alt', 'shift' ]
        
        return ( is_modifier, key_str, value )

    def get_keystrokes ( self, is_modifier : bool, key : str, value : int ) -> List[KeyStroke]:
        keystrokes : List[KeyStroke] = []

        ctrl = 'ctrl' in self.keyboard_state
        alt = 'alt' in self.keyboard_state
        shift = 'shift' in self.keyboard_state

        if is_modifier:
            for key in self.keyboard_state.keys():
                if key not in [ 'ctrl', 'alt', 'shift' ]:
                    keystrokes.append( KeyStroke( key, ctrl, alt, shift ) )
                    keystrokes.append( KeyStroke( value, ctrl, alt, shift ) )
        else:
            keystrokes.append( KeyStroke( key, ctrl, alt, shift ) )
            keystrokes.append( KeyStroke( value, ctrl, alt, shift ) )

        return keystrokes

    def on_press ( self, virtual_keyboard : KeyboardLibrary, key : Key ):
        ( is_modifier, key, value ) = self.get_key_info( key )

        self.keyboard_state[ key ] = True
        self.keyboard_state[ value ] = True

        keystrokes = self.get_keystrokes( is_modifier, key, value )
            
        virtual_keyboard.on_press( KeyEvent( key, value, True ) )

        for ks in keystrokes:
            if ks == KeyStroke( 'c', True, True, True ):
                raise KeyboardInterrupt()
            
            virtual_keyboard.on_press( ks )
            virtual_keyboard.on_press( KeyStrokePress( ks.key, ks.ctrl, ks.alt, ks.shift ) )

    def on_release ( self, virtual_keyboard : KeyboardLibrary, key : Key ):
        ( is_modifier, key, value ) = self.get_key_info( key )

        keystrokes = self.get_keystrokes( is_modifier, key, value )

        if key in self.keyboard_state:
            del self.keyboard_state[ key ]

        if value in self.keyboard_state:
            del self.keyboard_state[ value ]

        virtual_keyboard.on_press( KeyEvent( key, value, False ) )
        for ks in keystrokes:
            virtual_keyboard.on_release( ks )
            virtual_keyboard.on_press( KeyStrokeRelease( ks.key, ks.ctrl, ks.alt, ks.shift ) )

    def on_move ( self, virtual_keyboard : KeyboardLibrary, x : int, y : int ):
        virtual_keyboard.on_press( MouseMove( x, y ) )

    def on_click ( self, virtual_keyboard : KeyboardLibrary, x : int, y : int, button : int, pressed : bool ):
        virtual_keyboard.on_press( MouseClick( x, y, button, pressed ) )

    def on_scroll ( self, virtual_keyboard : KeyboardLibrary, x : int, y : int, dx : int, dy : int ):
        virtual_keyboard.on_press( MouseScroll( x, y, dx, dy ) )

    def listen ( self ):
        virtual_keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        if virtual_keyboard != None:
            loop = asyncio.get_running_loop()

            self.keyboard_listener = keyboard.Listener(
                on_press = lambda key: loop.call_soon_threadsafe( self.on_press, virtual_keyboard, key ),
                on_release = lambda key: loop.call_soon_threadsafe( self.on_release, virtual_keyboard, key ),
                suppress = False
            )

            self.keyboard_listener.start()

            # TODO: Conflicts with GLFW resizing and moving windows. Figure a solution
            # self.mouse_listener = mouse.Listener(
            #     on_move = lambda x, y: loop.call_soon_threadsafe( self.on_move, virtual_keyboard, x, y ),
            #     on_click = lambda x, y, button, pressed: loop.call_soon_threadsafe( self.on_click, virtual_keyboard, x, y, button, pressed ),
            #     on_scroll = lambda x, y, dx, dy: loop.call_soon_threadsafe( self.on_scroll, virtual_keyboard, x, y, dx, dy )
            # )

            # self.mouse_listener.start()

    def close ( self ):
        if self.keyboard_listener is not None:
            self.keyboard_listener.stop()

            self.keyboard_listener = None

        if self.mouse_listener is not None:
            self.mouse_listener.stop()

            self.mouse_listener = None

        


from musikla.core import Context
from musikla.core.events import MusicEvent
from musikla.core.events.transformers import Transformer
from typing import Any, Iterable, Optional
from configparser import ConfigParser
import argparse

class ArgumentParserError(Exception): pass

class ArgumentParser( argparse.ArgumentParser ):
    def error( self, message ):
        raise ArgumentParserError( message )

class Sequencer:
    def __init__ ( self ):
        self.realtime = False

        self.transformer : Optional[Transformer] = None

    def set_transformers ( self, *transformers : Transformer ):
        if transformers:
            self.transformer = Transformer.pipeline( 
                *transformers,
                Transformer.subscriber( self.on_event, self.on_close )
            )
        else:
            self.transformer = None

    @property
    def playing ( self ) -> bool:
        raise BaseException( "Abstract property Sequencer.playing accessed." )

    def get_time ( self ) -> int:
        raise BaseException( "Abstract method Sequencer.get_time called." )

    def on_event ( self, event : MusicEvent ):
        raise BaseException( "Abstract method Sequencer.on_event called." )
    
    def on_close ( self ):
        raise BaseException( "Abstract method Sequencer.on_close called." )

    def register_event ( self, event : MusicEvent ):
        if self.transformer is None:
            self.on_event( event )
        else:
            self.transformer.add_input( event )

    def register_events_many ( self, events : Iterable[MusicEvent] ):
        for event in events:
            self.register_event( event )

    def join ( self ):
        raise BaseException( "Abstract method Sequencer.join called." )

    def start ( self ):
        raise BaseException( "Abstract method Sequencer.start called." )

    def close ( self ):
        if self.transformer is None:
            self.on_close()
        else:
            self.transformer.end_input()

class SequencerFactory:
    default : bool = False

    def __init__ ( self, context : Context, config : ConfigParser ):
        self.name : str = ""
        self.argparser : Optional[ArgumentParser] = None
        self.context : Context = context
        self.config : ConfigParser = config

        self.init()
    
    def init ( self ):
        pass

    def from_str ( self, uri : str, args : Any = None ) -> Optional[Sequencer]:
        return None



from .enumerable import IterCursor, merge_sorted
from .context import Context, Library, StackFrame
from .shared_context import SharedContext
from .symbols_scope import SymbolsScope, Ref
from .instrument import Instrument, GeneralMidi
from .voice import Voice
from .value import Value, CallableValue
from .music import Music, TemplateMusic, MusicBuffer
from .clock import Clock
from .scheduler import Scheduler
from .metronome import Metronome


import time
from typing import Optional

class Clock():
    @staticmethod
    def epoch ():
        return Clock( auto_start = 0 )

    @staticmethod
    def _get_global_milliseconds () -> int:
        return int( round( time.time() * 1000 ) )

    def __init__ ( self, auto_start = True, start_time : int = None ):
        self.start_time : Optional[int] = None
        self.auto_start : bool = auto_start

    @property
    def started ( self ):
        return self.start_time is not None

    def start ( self ):
        self.start_time = self._get_global_milliseconds()

        return self

    def reset ( self ):
        self.start_time = None

        return self

    def elapsed ( self ):
        if self.auto_start and self.start_time is None:
            self.start()

        if self.start_time is None:
            return self._get_global_milliseconds()
        else:
            return self._get_global_milliseconds() - self.start_time



from musikla.libraries.keyboard.event import KeyboardEvent
from musikla.core import Context, Library
from musikla.core.theory import Note
from musikla.libraries.midi import MidiLibrary
from musikla.libraries.keyboard import KeyboardLibrary, EventSource, PianoKey
from colorama import Fore, Style
from typing import Any, Dict, Optional, cast
from mido.ports import BaseInput
import asyncio
import mido

class KeyboardMidoLibrary( Library ):
    def __init__ ( self ):
        super().__init__( "keyboard_mido" )

    def on_link ( self, script ):
        keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        if keyboard is not None:
            keyboard.add_source( KeyboardMidoEventSource( self.context ) )

        self.assign( 'MidiEvent', MidiEvent )

class KeyboardMidoEventSource( EventSource ):
    def __init__ ( self, context : Context ):
        self.context : Context = context
        self.midi_port : Optional[BaseInput] = None

    def on_message ( self, virtual_keyboard : KeyboardLibrary, msg : mido.Message ):
        if msg.type == 'note_on':
            note = Note().with_pitch( msg.note )

            virtual_keyboard.on_press( PianoKey( note ) )
        elif msg.type == 'note_off':
            note = Note().with_pitch( msg.note )

            virtual_keyboard.on_release( PianoKey( note ) )
        
        virtual_keyboard.on_press( MidiEvent( msg ) )

    def choose_port ( self, midi : MidiLibrary ):
        default_port = midi.get_midi_default_input()

        available_ports = mido.get_input_names()

        if default_port is not None and default_port not in available_ports:
            print( f"{ Fore.MAGENTA + 'WARN' + Style.RESET_ALL } Default selected midi port { Fore.MAGENTA + default_port + Style.RESET_ALL } is not available, ignoring." )
        
            default_port = None
        
        if default_port is None and available_ports:
            non_loopback_ports = [ p for p in available_ports if 'loopback' in p.lower() ]

            if non_loopback_ports:
                default_port = non_loopback_ports[ 0 ]
            else:
                default_port = available_ports[ 0 ]

            print( f"{ Fore.CYAN + 'INFO' + Style.RESET_ALL} No default selected midi port, using { Fore.CYAN + default_port + Style.RESET_ALL }." )

        return default_port

    def listen ( self ):
        virtual_keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )
        
        midi = cast( MidiLibrary, self.context.library( MidiLibrary ) )

        if virtual_keyboard is not None and midi is not None:
            loop = asyncio.get_running_loop()

            default_port = self.choose_port( midi ) #.get_midi_default_input()

            if default_port is not None:
                self.midi_port = mido.open_input( default_port, callback = lambda msg: loop.call_soon_threadsafe( self.on_message, virtual_keyboard, msg ) )

    def close ( self ):
        if self.midi_port is not None:
            self.midi_port.close()

            self.midi_port = None

class MidiEvent( KeyboardEvent ):
    binary : bool = False

    def __init__ ( self, message : mido.Message = None ):
        self.message : mido.Message = message
    
    def get_parameters ( self ) -> Dict[str, Any]:
        return { 'message': self.message }

    def __hash__ ( self ):
        return hash( '<MidiEvent>' )
    
    def __eq__ ( self, other ):
        if other is None:
            return False

        return isinstance( other, MidiEvent )



from fractions import Fraction
from typing import List, Dict

def keysToIndexes (d : Dict) -> Dict:
    i : int = 0
    result = dict()

    for key in d.keys():
        result[ key ] = i
        i += 1

    return result

class NoteAccidental():
    DOUBLEFLAT = -2
    FLAT = -1
    NONE = 0
    SHARP = 1
    DOUBLESHARP = 2

NotePitchClasses = {
    'C': 0,
    'D': 2,
    'E': 4,
    'F': 5,
    'G': 7,
    'A': 9,
    'B': 11
}

NotePitchClassesInv : Dict[int, str] = { v: k for k, v in NotePitchClasses.items() }

NotePitchClassesIndexes : Dict[str, int] = keysToIndexes( NotePitchClasses )

class Note:
    # Static
    @staticmethod
    def parse_pitch_octave ( pitch ):
        if pitch[ 0 ].islower():
            pitch_class = NotePitchClasses[ pitch[ 0 ].upper() ]
            octave = len( pitch ) - 1

            return ( pitch_class, octave )
        else:
            pitch_class = NotePitchClasses[ pitch[ 0 ] ]
            octave = -len( pitch )

            return ( pitch_class, octave )

    @staticmethod
    def from_pitch ( pitch : int, value : Fraction = Fraction() ):
        octave = ( pitch // 12 ) - 1
        pitch_class = pitch % 12
        accidental = 0

        if pitch_class not in NotePitchClassesInv:
            pitch_class -= 1

            accidental = 1

        return Note( pitch_class, octave, accidental, value )

    def __init__ ( self, pitch_class : int = 0, octave : int = 0, accidental : int = 0, value : Fraction = Fraction() ):
        self.pitch_class : int = pitch_class
        self.octave : int = octave
        self.accidental : int = accidental
        self.value : Fraction = value

    def to_pitch ( self ) -> int:
        return ( self.octave + 1 ) * 12 + self.pitch_class + self.accidental

    def with_pitch ( self, pitch : int ) -> 'Note':
        self.octave = ( pitch // 12 ) - 1
        self.pitch_class = pitch % 12

        if self.pitch_class not in NotePitchClassesInv:
            self.pitch_class -= 1

            self.accidental = 1
        else:
            self.accidental = 0

        return self

    def transpose ( self, semitones : int ) -> 'Note':
        return self.with_pitch( self.to_pitch() + int( semitones ) )

    def timeless ( self ):
        if self.value != None and self.value != 1:
            note = self.clone()
            note.value = None
            return note
        else:
            return self

    def clone ( self ):
        return Note( self.pitch_class, self.octave, self.accidental, self.value )

    def as_chord ( self, chord : List[int] ) -> 'Chord':
        return Chord( self, chord )

    def __eq__ ( self, other ):
        if other is None:
            return False

        return int( self ) == int( other )

    def __int__ ( self ):
        return self.to_pitch()

    def __hash__ ( self ):
        return hash( str( self ) )

    def to_string ( self, base_octave : int = 4, append_value : bool = True ):
        note : str = NotePitchClassesInv[ self.pitch_class ].lower()

        if self.octave < base_octave:
            note = note.upper()

            for _ in range( base_octave - 2, self.octave - 1, -1 ): note += ","
        else:
            for _ in range( base_octave + 1, self.octave + 1 ): note += "'"

        if append_value and self.value != None and self.value != 1:
            note += str( Fraction( self.value ) )

        if self.accidental == NoteAccidental.DOUBLESHARP:
            note = '^^' + note
        elif self.accidental == NoteAccidental.SHARP:
            note = '^' + note
        elif self.accidental == NoteAccidental.FLAT:
            note = '_' + note
        elif self.accidental == NoteAccidental.DOUBLEFLAT:
            note = '__' + note

        return note

    def __str__ ( self ):
        return self.to_string()



from musikla.parser.printer import CodePrinter
from .expression_node import ExpressionNode
from musikla.core import Value
from typing import List, Tuple

class Object(dict):
    def __init__ ( self, pairs = [] ):
        super().__init__( pairs )
    
    def has ( self, key ):
        return key in self
    
    def get ( self, key, default = None ):
        if key in self:
            return self[ key ]

        return default
        
    def set ( self, key, value ):
        self[ key ] = value

        return self
        
    def delete ( self, key ):
        del self[ key ]

    def __delattr__ ( self, key ):
        del self[ key ]

    def __getattr__ ( self, key ):
        return self[ key ]

    def __setattr__ ( self, key, value ):
        self[ key ] = value

class ObjectLiteralNode( ExpressionNode ):
    def __init__ ( self, values : List[Tuple[str, ExpressionNode]], position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.values : List[Tuple[str, ExpressionNode]] = values

    def __eval__ ( self, context ):
        return Object( [ ( key, Value.assignment( Value.eval( context.fork(), node ) ) ) for key, node in self.values ] )

    def to_source ( self, printer : CodePrinter ):
        with printer.block( '@{', '}' ):
            for i in range( len( self.values ) ):
                if i > 0:
                    printer.add_token( '; ' )
                
                key, node = self.values[ i ]

                printer.add_token( str( key ) + ' = ' )

                node.to_source( printer )            



from musikla.core.music import MusicBuffer
from .player import PlayerLike
from musikla.core.events import NoteOffEvent, ChordOffEvent
from musikla.core.events.transformers import DecomposeNotesTransformer, BalanceNotesTransformer
from typing import Callable, List, Optional
from musikla.parser.abstract_syntax_tree import Node
from asyncio import Future, sleep, wait, FIRST_COMPLETED

class InteractivePlayer:
    def __init__ ( self, factory : Callable, player : PlayerLike, start_time : int = 0, repeat : bool = False, extend : bool = False, realtime : bool = False ):
        self.factory : Callable = factory
        self.player : PlayerLike = player
        self.buffers : Optional[List[MusicBuffer]] = None
        self.repeat : bool = repeat
        self.extend : bool = extend

        # How many milliseconds to try and buffer, minimum
        self.buffer_duration : int = 50

        self.extended_notes = []
        self.events_iterator = None
        self.is_playing : bool = False
        self.realtime : bool = realtime
        self.cancel_token : Optional[CancelToken] = None

    def start_gen ( self ):
        cancel_token = self.cancel_token or CancelToken()
        try:
            for events_iterator in flat_factory( self.factory ):
                self.is_playing = True

                # First make sure all note events are separated into NoteOn and NoteOff
                events_iterator = DecomposeNotesTransformer.iter( events_iterator )
                # When extending the notes, ignore any early NoteOn events
                if self.extend: 
                    events_iterator = filter( lambda ev: not isinstance( ev, NoteOffEvent ) and not isinstance( ev, ChordOffEvent ), events_iterator )
                # Finally append any missing NoteOff's that might have been cut off because the player stopped playing
                events_iterator = cancellable( events_iterator, cancel_token )

                events_iterator = BalanceNotesTransformer.iter( events_iterator, self.player.get_time )

                for event in events_iterator:
                    if self.extend and ( isinstance( event, NoteOffEvent ) or isinstance( event, ChordOffEvent ) ):
                        self.extended_notes.append( event )
                    else:
                        yield event

                if cancel_token.cancelled or not self.repeat:
                    break
        finally:
            if not self.extend:
                self.is_playing = False

                self.cancel_token = None
    
    def stop_gen ( self ):
        if self.cancel_token is not None:
            self.cancel_token.cancel()

    def start_sync ( self ):
        if self.is_playing:
            return

        self.cancel_token = CancelToken()

        self.player.play_more( self.start_gen() )
    
    def stop_sync ( self, now : int ):
        def extended_gen ():
            try:
                for ev in self.extended_notes:
                    yield ev.clone( timestamp = now )
            finally:
                self.is_playing = False
            
        self.stop_gen()
        
        if self.extend:
            self.player.play_more( extended_gen() )
        else:
            self.is_playing = False


    async def start ( self ):
        if self.is_playing:
            return

        self.cancel_token = CancelToken()

        sync_iterator = self.start_gen()

        if self.realtime:
            iterator = realtime( sync_iterator, self.cancel_token.future, lambda e: e.timestamp, self.player.get_time, self.buffer_duration )
        else:
            iterator = iter_to_aiter( sync_iterator )

        try:
            async for event in iterator:
                self.player.play_more( [ event ] )

                if self.buffers is not None:
                    for buf in self.buffers: 
                        buf.append( event )
        except BaseException as ex:
            self.player.print_error( ex )

        if self.extend and self.extended_notes:
            await self.cancel_token.future

            now = self.player.get_time()

            events = [ ev.clone( timestamp = now ) for ev in self.extended_notes ]

            self.player.play_more( events )

            if self.buffers is not None:
                for buf in self.buffers: 
                    buf.extend( events )

        self.is_playing = False

    async def stop ( self ):
        self.stop_gen()
    
async def wait_first ( self, *aws ):
    done, _ = await wait( aws, return_when = FIRST_COMPLETED )

    for item in done:
        return item.result()


async def realtime ( source, stop_future, get_timestamp : Callable, get_time : Callable, offset : int ):
    time = get_time()

    for event in source:
        timestamp = get_timestamp( event )

        if timestamp > time - offset:
            stopped = await wait_first( stop_future, sleep( ( timestamp - time - offset ) / 1000, result = False ) )

            if stopped: break

            time = get_time()

        if stop_future.done() and stop_future.result(): break

        yield event
        
async def take_while ( source, predicate : Callable ):
    async for event in source:
        if not predicate(event):
            break
        
        yield event
        
def flat_factory ( factory, count : int = None ):
    i = 0

    while True:
        if count is not None:
            if i >= count:
                break

            i += 1

        iterable = factory()

        if iterable != None and hasattr( iterable, '__iter__' ):
            yield iterable
        else:
            break

class CancelToken():
    def __init__ ( self ):
        self.cancelled = False
        self.future : Future[bool] = Future()
    
    def cancel ( self ):
        if not self.future.done():
            self.cancelled = True
            self.future.set_result( True )

def cancellable ( it, cancel_token : CancelToken ):
    if cancel_token.cancelled:
        return

    for event in it:
        if cancel_token.cancelled:
            break

        yield event

async def iter_to_aiter ( it ):
    for item in it:
        yield item
